import os
import tempfile

import h5py
import numpy as np
import pytest

import treams
from treams import io


@pytest.mark.gmsh
def test_meshspheres():
    import gmsh

    gmsh.initialize()
    gmsh.model.add("spheres")
    io.mesh_spheres([1, 2], [[0, 0, 2], [0, 0, -2]], gmsh.model)
    with tempfile.TemporaryDirectory() as directory:
        filename = os.path.join(directory, "spheres.msh")
        gmsh.write(filename)
        gmsh.finalize()
        with open(filename, "r") as f:
            value = "".join([line for line, _ in zip(f, range(5))][3:])
    expect = "$Entities\n4 6 2 2\n"
    assert value == expect


class TestSaveHDF5:
    def test_helicity(self):
        with h5py.File("test.h5", "x", driver="core", backing_store=False) as fp:
            m = np.arange(4 * 3 * 16 * 16).reshape((4, 3, 16, 16))
            tms = [
                [
                    treams.TMatrix(
                        m[i, ii], k0=7, material=treams.Material(i + 1, ii + 1, 0.5)
                    )
                    for ii in range(3)
                ]
                for i in range(4)
            ]
            io.save_hdf5(
                fp,
                tms,
                "testname",
                "testdescr",
                lunit="µm",
                uuid=b"12345678123456781234567812345678",
            )
            assert np.all(fp["tmatrix"] == m)
            assert fp["angular_vacuum_wavenumber"][...] == 7
            assert np.all(
                fp["materials/embedding/relative_permittivity"]
                == np.arange(1, 5)[:, None] * [1, 1, 1]
            )
            assert np.all(
                fp["materials/embedding/relative_permeability"] == np.arange(1, 4)
            )
            assert fp["materials/embedding/chirality"][...] == 0.5
            assert fp["embedding"] == fp["materials/embedding"]
            assert np.all(fp["modes/l"] == tms[0][0].basis.l)
            assert np.all(fp["modes/m"] == tms[0][0].basis.m)
            assert np.all(
                [i.decode() for i in fp["modes/polarization"][...]]
                == 8 * ["positive", "negative"]
            )
            assert fp["uuid"][()] == b"12345678123456781234567812345678"
            assert fp.attrs["name"] == "testname"
            assert fp.attrs["description"] == "testdescr"
            assert fp["angular_vacuum_wavenumber"].attrs["unit"] == r"µm^{-1}"

    def test_parity(self):
        with h5py.File("test.h5", "x", driver="core", backing_store=False) as fp:
            m = np.arange(4 * 3 * 16 * 16).reshape((4, 3, 16, 16))
            tms = [
                [
                    treams.TMatrix(
                        m[i, ii], k0=i + 1, material=treams.Material(), poltype="parity"
                    )
                    for ii in range(3)
                ]
                for i in range(4)
            ]
            io.save_hdf5(fp, tms, "testname", "testdescr")
            assert (
                np.all(fp["tmatrix"] == m)
                and np.all(fp["angular_vacuum_wavenumber"] == np.arange(1, 5)[:, None])
                and fp["materials/embedding/relative_permittivity"][...] == 1
                and fp["materials/embedding/relative_permeability"][...] == 1
                and fp["embedding"] == fp["materials/embedding"]
                and np.all(fp["modes/l"] == tms[0][0].basis.l)
                and np.all(fp["modes/m"] == tms[0][0].basis.m)
                and np.all(
                    [i.decode() for i in fp["modes/polarization"][...]]
                    == 8 * ["electric", "magnetic"]
                )
            )


class TestLoadHdf5:
    def test(self):
        with h5py.File("test.h5", "x", driver="core", backing_store=False) as fp:
            fp.create_dataset(
                "tmatrix", data=np.arange(4 * 3 * 6 * 4).reshape((4, 3, 6, 4))
            )
            fp.create_dataset("frequency", data=np.arange(1, 5)[:, None])
            fp["frequency"].attrs["unit"] = "THz"
            fp.create_dataset("materials/foo/refractive_index", data=4)
            fp["embedding"] = h5py.SoftLink("/materials/foo")
            fp.create_dataset("modes/positions", data=[[0, 0, 0]])
            fp.create_dataset("modes/l_incident", data=[1, 1, 2, 2])
            fp.create_dataset("modes/l_scattered", data=6 * [1])
            fp.create_dataset("modes/m_incident", data=4 * [0])
            fp.create_dataset("modes/m_scattered", data=[-1, -1, 0, 0, 1, 1])
            fp.create_dataset(
                "modes/polarization_incident", data=2 * ["electric", "magnetic"]
            )
            fp.create_dataset(
                "modes/polarization_scattered", data=3 * ["electric", "magnetic"]
            )

            tms = io.load_hdf5(fp)

        basis = treams.SphericalWaveBasis(
            [
                (1, 0, 0),
                (1, 0, 1),
                (2, 0, 0),
                (2, 0, 1),
                (1, -1, 0),
                (1, -1, 1),
                (1, 1, 0),
                (1, 1, 1),
            ]
        )
        assert tms.shape == (4, 3)
        assert abs(tms[0, 0].k0 - 2 * np.pi / 299792.458) < 1e-16
        assert abs(tms[0, 1].k0 - 2 * np.pi / 299792.458) < 1e-16
        assert abs(tms[1, 0].k0 - 4 * np.pi / 299792.458) < 1e-16
        assert tms[0, 0].poltype == "parity"
        assert tms[0, 0].basis <= basis and basis <= tms[0, 0].basis
        assert np.all(
            tms[3, 2]
            == [
                [272, 273, 274, 275, 0, 0, 0, 0],
                [276, 277, 278, 279, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0],
                [264, 265, 266, 267, 0, 0, 0, 0],
                [268, 269, 270, 271, 0, 0, 0, 0],
                [280, 281, 282, 283, 0, 0, 0, 0],
                [284, 285, 286, 287, 0, 0, 0, 0],
            ]
        )



import copy

import numpy as np

import treams
from treams import TMatrix


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class TestInit:
    def test_simple(self):
        tm = TMatrix(np.eye(6), k0=1)
        assert (
            np.all(tm == np.eye(6))
            and tm.k0 == 1
            and tm.material == (1, 1, 0)
            and tm.basis == treams.SphericalWaveBasis.default(1)
            and tm.poltype == treams.config.POLTYPE
        )

    def test_complex(self):
        tm = TMatrix(
            np.diag([1, 2]),
            k0=3,
            material=[2, 8, 1],
            basis=treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]], [1, 0, 0]),
        )
        assert (
            np.all(tm == np.diag([1, 2]))
            and tm.k0 == 3
            and tm.material == (2, 8, 1)
            and tm.basis == treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]], [1, 0, 0])
        )


class TestSphere:
    def test(self):
        tm = TMatrix.sphere(2, 3, 4, [(2, 1, 1), (9, 1, 2)])
        del tm.modetype
        m = treams.coeffs.mie([1, 2], [12], [2, 9], [1, 1], [1, 2])
        assert (
            np.all(np.diag(tm)[:6:2] == m[0, 1, 1])
            and np.all(np.diag(tm)[1:6:2] == m[0, 0, 0])
            and np.all(np.diag(tm)[6::2] == m[1, 1, 1])
            and np.all(np.diag(tm)[7::2] == m[1, 0, 0])
            and np.all(np.diag(tm, -1)[:6:2] == m[0, 0, 1])
            and np.all(np.diag(tm, 1)[:6:2] == m[0, 1, 0])
            and np.all(np.diag(tm, -1)[6::2] == m[1, 0, 1])
            and np.all(np.diag(tm, 1)[6::2] == m[1, 1, 0])
            and np.all(np.diag(tm, -1)[1:6:2] == 0)
            and np.all(np.diag(tm, 1)[1:6:2] == 0)
            and np.all(np.diag(tm, -1)[7::2] == 0)
            and np.all(np.diag(tm, 1)[7::2] == 0)
            and np.all(np.triu(tm, 2) == 0)
            and np.all(np.tril(tm, -2) == 0)
            and tm.k0 == 3
            and tm.material == (9, 1, 2)
            and tm.poltype == "helicity"
            and tm.basis == treams.SphericalWaveBasis.default(2)
        )


class TestProperties:
    def test_xs_ext_avg(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.xs_ext_avg, 2.9294930236877077)

    def test_xs_ext_avg_kappa_zero(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 3), (9, 4)])
        assert isclose(tm.xs_ext_avg, 0.15523595021864234)

    def test_xs_sca_avg(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.xs_sca_avg, 1.6603264386283758)

    def test_xs_sca_avg_kappa_zero(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 3), (9, 4)])
        assert isclose(tm.xs_sca_avg, 0.08434021223849283)

    def test_cd(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.cd, -0.9230263013362784)

    def test_chi(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.chi, 0.7483463517622965)

    def test_db(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.db, 0.6085814346536764)

    def test_modes(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        l, m, pol = tm.basis.lms
        assert (
            np.all(l == 6 * [1] + 10 * [2])
            and np.all(m == [-1, -1, 0, 0, 1, 1, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2])
            and np.all(pol == [int((i + 1) % 2) for i in range(16)])
        )


class TestXs:
    def test(self):
        tm = TMatrix.sphere(2, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        illu = treams.plane_wave([0, 0, 1], [0.5, 0], k0=tm.k0, material=tm.material)
        xs = tm.xs(illu, 0.125)
        assert isclose(xs[0], 3.194830855171616,) and isclose(xs[1], 5.63547158)


class TestTranslate:
    def test(self):
        tm = TMatrix.sphere(3, 0.1, [0.2], [(2 + 1j, 1.1, 1), (9, 1, 2)])
        m = copy.deepcopy(tm)
        rs = np.array([[0.1, 0.2, 0.3], [-0.4, -0.5, -0.4]])
        tm = tm.translate(rs[0])
        tm = tm.translate(rs[1])
        tm = tm.translate(-rs[0] - rs[1])
        assert np.all(np.abs(tm - m) < 1e-8)

    def test_kappa_zero(self):
        tm = TMatrix.sphere(3, 0.1, [0.2], [(2 + 1j, 1.1), (9, 1)])
        m = copy.deepcopy(tm)
        rs = np.array([[0.1, 0.2, 0.3], [-0.4, -0.5, -0.4]])
        tm.translate(rs[0])
        tm.translate(rs[1])
        tm.translate(-rs[0] - rs[1])
        assert np.all(np.abs(tm - m) < 1e-8)


class TestClusterRotate:
    def test(self):
        tms = [TMatrix.sphere(3, 0.1, [0.1], [i * i, 1]) for i in range(1, 5)]
        rs1 = np.array([[0, 0, 0], [0.2, 0, 0], [0, 0.2, 0], [0, 0, 0.2]])
        tm1 = TMatrix.cluster(tms, rs1)
        tm1 = tm1.interaction.solve().expand(treams.SphericalWaveBasis.default(3))
        tm1 = tm1.rotate(1, 2, 3)
        a = np.array([[np.cos(1), -np.sin(1), 0], [np.sin(1), np.cos(1), 0], [0, 0, 1]])
        b = np.array([[np.cos(2), 0, np.sin(2)], [0, 1, 0], [-np.sin(2), 0, np.cos(2)]])
        c = np.array([[np.cos(3), -np.sin(3), 0], [np.sin(3), np.cos(3), 0], [0, 0, 1]])
        rs2 = (a @ b @ c @ rs1.T).T
        tm2 = TMatrix.cluster(tms, rs2)
        tm2 = tm2.interaction.solve().expand(treams.SphericalWaveBasis.default(3))
        assert np.all(np.abs(tm1 - tm2) < 1e-16)



import numpy as np

from treams import pw


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class TestToCw:
    def test(self):
        assert isclose(pw.to_cw(1, 4, 0, 3, 2, 1, 0), np.power((2 + 3j), 4) / 169)

    def test_complex(self):
        assert isclose(pw.to_cw(1, 4, 0, 3, 2j, 1, 0), 25)

    def test_one(self):
        assert pw.to_cw(1, 0, 0, 3, 2, 1, 0) == 1

    def test_pol(self):
        assert pw.to_cw(5, 4, 1, 3, 2, 1, 0) == 0


class TestToSw:
    def test_h(self):
        assert isclose(
            pw.to_sw(5, 4, 0, 3, 2, 1, 0), -3.6625822219731057 + 3.632060703456664j
        )

    def test_h_zero(self):
        assert pw.to_sw(5, 4, 1, 3, 2, 1j, 0) == 0

    def test_p_same(self):
        assert isclose(
            pw.to_sw(5, 4, 0, 3, 2, 1, 0, poltype="parity"),
            -1.3753807114589616 + 1.3639192055301372j,
        )

    def test_p_opposite(self):
        assert isclose(
            pw.to_sw(5, 4, 1, 3, 2, 1j, 0, poltype="parity"),
            3.087192594399995 + 3.1131353893109184j,
        )


class TestTranslate:
    def test(self):
        assert pw.translate(1, 2, 3, 4, 5, 6) == np.exp(32j)


class TestPermuteXyz:
    def test_h(self):
        assert isclose(
            pw.permute_xyz(1, 2, 3, 0, 0), (-6 + 1j * np.sqrt(14)) / np.sqrt(50)
        )

    def test_h_inv(self):
        assert isclose(
            pw.permute_xyz(1, 2, 1j, 0, 0, inverse=True), -1j * np.sqrt(5 / 3)
        )

    def test_h_opposite(self):
        assert pw.permute_xyz(1, 2, 3, 0, 1) == 0

    def test_h_kxy_zero(self):
        assert pw.permute_xyz(0, 0, -1, 0, 0) == 1j

    def test_p(self):
        assert isclose(
            pw.permute_xyz(1, 2, 3, 0, 0, poltype="parity"), -6 / np.sqrt(50)
        )

    def test_p_inv(self):
        assert pw.permute_xyz(
            1, 2, 1j, 0, 0, poltype="parity", inverse=True
        ) == -1j / np.sqrt(15)

    def test_p_opposite(self):
        assert pw.permute_xyz(1, 2, 3, 0, 1, poltype="parity") == -1j * np.sqrt(
            14
        ) / np.sqrt(50)

    def test_p_inv_opposite(self):
        assert pw.permute_xyz(
            1, 2, 1j, 0, 1, poltype="parity", inverse=True
        ) == 4j / np.sqrt(15)

    def test_p_kxy_zero(self):
        assert pw.permute_xyz(0, 0, -1, 0, 0, poltype="parity") == 0

    def test_p_kxy_zero_opposite(self):
        assert pw.permute_xyz(0, 0, -1, 0, 1, poltype="parity") == -1j



"""Sphinx config."""
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from importlib.metadata import version

import numpy as np

# sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = "treams"
copyright = "2023, Dominik Beutel"
author = "Dominik Beutel"
release = version("treams")
version = ".".join(release.split(".")[:3])


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "matplotlib.sphinxext.plot_directive",
]

autosectionlabel_prefix_document = True
autosummary_filename_map = {
    "treams.Lattice": "treams.Lattice_class",
    "treams.Lattice.reciprocal": "treams.Lattice_class.reciprocal",
    "treams.Lattice.volume": "treams.Lattice_class.volume",
}
doctest_global_setup = """
import numpy as np
from treams import *

np.set_printoptions(precision=3, suppress=True)
"""
intersphinx_mapping = {
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "numpy": ("https://numpy.org/doc/stable", None),
}
todo_include_todos = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pyramid"
html_css_files = ["custom.css"]
html_sidebars = {
    "**": ["globaltoc.html", "relations.html", "sourcelink.html", "searchbox.html"]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def preprocess_ufunc(app, what, name, obj, options, lines):
    """Remove the first four docstring lines from numpy ufuncs.

    The docstring of ufuncs have a line containing the function name and signature and a
    blank line automatically prepended. The signature contains all arguments and keyword
    arguments but the positional arguments only get dummy variable names ('x1', 'x2',
    ...) and is therefore often confusing.

    By convention all ufuncs contain in their first manually added lines the function
    name with a simplified signature followed by a blank line. We explicitly add the
    signature with a separate process to autodoc.

    The four lines containing function names and signatures or being blank are therefore
    removed for a clean documentation produced by sphinx.
    """
    if isinstance(obj, np.ufunc):
        lines.pop(0)
        lines.pop(0)
        lines.pop(0)
        lines.pop(0)


def ufunc_signature(app, what, name, obj, options, signature, return_annotation):
    """Extract the signature from the docstring for numpy ufuncs.

    As described in :func:`preprocess_ufunc` the third line in a ufunc docstring
    contains the manually added function name and signature intended to be displayed in
    the documentation (the preceeding two lines are automatically added). We extract the
    signature and add it manually.
    """
    if isinstance(obj, np.ufunc):
        signature = "(" + obj.__doc__.split("\n")[2].split("(")[1]
    return signature, return_annotation


def setup(app):
    """Add processes to autodoc."""
    app.connect("autodoc-process-docstring", preprocess_ufunc)
    app.connect("autodoc-process-signature", ufunc_signature)



import matplotlib.pyplot as plt
import numpy as np

import treams

k0 = 2 * np.pi / 700
materials = [treams.Material(-16.5 + 1j), treams.Material()]
lmax = 3
radii = [75, 75]
positions = [[-75, 0, 30], [75, 0, -30]]
lattice = treams.Lattice.square(300)
kpar = [0, 0]

spheres = [treams.TMatrix.sphere(lmax, k0, r, materials) for r in radii]
chain = treams.TMatrix.cluster(spheres, positions).latticeinteraction.solve(
    lattice, kpar
)

inc = treams.plane_wave([0, 0, k0], [-1, 0, 0], k0=chain.k0, material=chain.material)
sca = chain @ inc.expand(chain.basis)

grid = np.mgrid[-150:150:31j, 0:1, -150:150:31j].squeeze().transpose((1, 2, 0))
ez = np.zeros_like(grid[..., 0])
valid = chain.valid_points(grid, radii)
vals = []
for i, r in enumerate(grid[valid]):
    swb = treams.SphericalWaveBasis.default(1, positions=[r])
    field = sca.expandlattice(basis=swb).efield(r)
    vals.append(np.real(inc.efield(r)[2] + field[2]))
ez[valid] = vals
ez[~valid] = np.nan

fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    grid[0, :, 2], grid[:, 0, 0], ez.T, shading="nearest", vmin=-1, vmax=1,
)
cb = plt.colorbar(pcm)
cb.set_label("$E_z$")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
fig.show()



import numpy as np

import treams
from treams import TMatrixC


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class TestInit:
    def test_simple(self):
        tm = TMatrixC(np.eye(6), k0=1)
        assert (
            np.all(tm == np.eye(6))
            and tm.k0 == 1
            and tm.material.epsilon == 1
            and tm.material.mu == 1
            and tm.material.kappa == 0
            and np.all(tm.basis.positions == [[0, 0, 0]])
            and tm.poltype == "helicity"
            and np.all(tm.basis.kz == 6 * [0])
            and np.all(tm.basis.m == [-1, -1, 0, 0, 1, 1])
            and np.all(tm.basis.pol == [1, 0, 1, 0, 1, 0])
            and np.all(tm.basis.pidx == [0, 0, 0, 0, 0, 0])
            and np.all(tm.ks == [1, 1])
        )

    def test_complex(self):
        tm = TMatrixC(
            np.diag([1, 2]),
            k0=3,
            material=treams.Material(2, 8, 1),
            basis=treams.CylindricalWaveBasis([[1, 0, 0], [1, 0, 1]], [1, 0, 0]),
        )
        assert (
            np.all(tm == np.diag([1, 2]))
            and tm.k0 == 3
            and tm.material.epsilon == 2
            and tm.material.mu == 8
            and tm.material.kappa == 1
            and np.all(tm.basis.positions == [[1, 0, 0]])
            and tm.poltype == "helicity"
            and np.all(tm.basis.kz == [1, 1])
            and np.all(tm.basis.m == [0, 0])
            and np.all(tm.basis.pol == [0, 1])
            and np.all(tm.basis.pidx == [0, 0])
            and np.all(tm.ks == [9, 15])
        )


class TestCylinder:
    def test(self):
        tm = TMatrixC.cylinder([1], 2, 3, 4, [(2, 1, 1), (9, 1, 2)])
        m = treams.coeffs.mie_cyl(1, [-2, -1, 0, 1, 2], 3, [4], [2, 9], [1, 1], [1, 2])
        assert (
            tm[0, 0] == m[0, 1, 1]
            and tm[1, 1] == m[0, 0, 0]
            and tm[0, 1] == m[0, 1, 0]
            and tm[1, 0] == m[0, 0, 1]
            and tm[2, 2] == m[1, 1, 1]
            and tm[3, 3] == m[1, 0, 0]
            and tm[2, 3] == m[1, 1, 0]
            and tm[3, 2] == m[1, 0, 1]
            and tm[4, 4] == m[2, 1, 1]
            and tm[5, 5] == m[2, 0, 0]
            and tm[4, 5] == m[2, 1, 0]
            and tm[5, 4] == m[2, 0, 1]
            and tm[6, 6] == m[3, 1, 1]
            and tm[7, 7] == m[3, 0, 0]
            and tm[6, 7] == m[3, 1, 0]
            and tm[7, 6] == m[3, 0, 1]
            and tm[8, 8] == m[4, 1, 1]
            and tm[9, 9] == m[4, 0, 0]
            and tm[8, 9] == m[4, 1, 0]
            and tm[9, 8] == m[4, 0, 1]
            and tm.k0 == 3
            and tm.material.epsilon == 9
            and tm.material.mu == 1
            and tm.material.kappa == 2
            and np.all(tm.basis.positions == [[0, 0, 0]])
            and tm.poltype == "helicity"
            and np.all(tm.basis.kz == 10 * [1])
            and np.all(tm.basis.m == [-2, -2, -1, -1, 0, 0, 1, 1, 2, 2])
            and np.all(tm.basis.pol == [int((i + 1) % 2) for i in range(10)])
            and np.all(tm.basis.pidx == 10 * [0])
            and np.all(tm.ks == [3, 15])
        )


class TestProperties:
    def test_xw_ext_avg(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.xw_ext_avg, 1.2047234457348013)

    def test_xw_ext_avg_kappa_zero(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 3), (9,)])
        assert isclose(tm.xw_ext_avg, 0.6661748147466017)

    def test_xw_sca_avg(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        assert isclose(tm.xw_sca_avg, 0.6847986332271352)

    def test_xw_sca_avg_kappa_zero(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 3), (9, 4)])
        assert isclose(tm.xw_sca_avg, 0.18067830829683562)

    def test_krho(self):
        tm = TMatrixC.cylinder([0, 5], 1, 3, [1], [(2 + 1j,), ()])
        assert np.all(
            np.abs(tm.krhos - [3, 3, 3, 3, 3, 3, 4j, 4j, 4j, 4j, 4j, 4j]) < 1e-16
        )

    def test_modes(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        kz, m, pol = tm.basis.zms
        assert (
            np.all(kz == 6 * [-1] + 6 * [1])
            and np.all(m == [-1, -1, 0, 0, 1, 1, -1, -1, 0, 0, 1, 1])
            and np.all(pol == [int((i + 1) % 2) for i in range(12)])
        )

    def test_fullmodes(self):
        tm = TMatrixC.cylinder([-1, 1], 1, 3, [4], [(2 + 1j, 1, 1), (9, 1, 2)])
        pidx, kz, m, pol = tm.basis[()]
        assert (
            np.all(kz == 6 * [-1] + 6 * [1])
            and np.all(m == [-1, -1, 0, 0, 1, 1, -1, -1, 0, 0, 1, 1])
            and np.all(pol == [int((i + 1) % 2) for i in range(12)])
            and np.all(pidx == 12 * [0])
        )



"""Loading and storing data.

Most functions rely on at least one of the external packages `h5py` or `gmsh`.
"""

import sys
import uuid as _uuid
from importlib.metadata import version

import numpy as np

import treams

try:
    import h5py
except ImportError:
    h5py = None

LENGTHS = {
    "ym": 1e-24,
    "zm": 1e-21,
    "am": 1e-18,
    "fm": 1e-15,
    "pm": 1e-12,
    "nm": 1e-9,
    "um": 1e-6,
    "µm": 1e-6,
    "mm": 1e-3,
    "cm": 1e-2,
    "dm": 1e-1,
    "m": 1,
    "dam": 1e1,
    "hm": 1e2,
    "km": 1e3,
    "Mm": 1e6,
    "Gm": 1e9,
    "Tm": 1e12,
    "Pm": 1e15,
    "Em": 1e18,
    "Zm": 1e21,
    "Ym": 1e24,
}

INVLENGTHS = {
    r"ym^{-1}": 1e24,
    r"zm^{-1}": 1e21,
    r"am^{-1}": 1e18,
    r"fm^{-1}": 1e15,
    r"pm^{-1}": 1e12,
    r"nm^{-1}": 1e9,
    r"um^{-1}": 1e6,
    r"µm^{-1}": 1e6,
    r"mm^{-1}": 1e3,
    r"cm^{-1}": 1e2,
    r"dm^{-1}": 1e1,
    r"m^{-1}": 1,
    r"dam^{-1}": 1e-1,
    r"hm^{-1}": 1e-2,
    r"km^{-1}": 1e-3,
    r"Mm^{-1}": 1e-6,
    r"Gm^{-1}": 1e-9,
    r"Tm^{-1}": 1e-12,
    r"Pm^{-1}": 1e-15,
    r"Em^{-1}": 1e-18,
    r"Zm^{-1}": 1e-21,
    r"Ym^{-1}": 1e-24,
}

FREQUENCIES = {
    "yHz": 1e-24,
    "zHz": 1e-21,
    "aHz": 1e-18,
    "fHz": 1e-15,
    "pHz": 1e-12,
    "nHz": 1e-9,
    "uHz": 1e-6,
    "µHz": 1e-6,
    "mHz": 1e-3,
    "cHz": 1e-2,
    "dHz": 1e-1,
    "s": 1,
    "daHz": 1e1,
    "hHz": 1e2,
    "kHz": 1e3,
    "MHz": 1e6,
    "GHz": 1e9,
    "THz": 1e12,
    "PHz": 1e15,
    "EHz": 1e18,
    "ZHz": 1e21,
    "YHz": 1e24,
    r"ys^{-1}": 1e24,
    r"zs^{-1}": 1e21,
    r"as^{-1}": 1e18,
    r"fs^{-1}": 1e15,
    r"ps^{-1}": 1e12,
    r"ns^{-1}": 1e9,
    r"us^{-1}": 1e6,
    r"µs^{-1}": 1e6,
    r"ms^{-1}": 1e3,
    r"cs^{-1}": 1e2,
    r"ds^{-1}": 1e1,
    r"s^{-1}": 1,
    r"das^{-1}": 1e-1,
    r"hs^{-1}": 1e-2,
    r"ks^{-1}": 1e-3,
    r"Ms^{-1}": 1e-6,
    r"Gs^{-1}": 1e-9,
    r"Ts^{-1}": 1e-12,
    r"Ps^{-1}": 1e-15,
    r"Es^{-1}": 1e-18,
    r"Zs^{-1}": 1e-21,
    r"Ys^{-1}": 1e-24,
}


def mesh_spheres(radii, positions, model, meshsize=None, meshsize_boundary=None):
    """Generate a mesh of multiple spheres.

    This function facilitates generating a mesh for a cluster spheres using gmsh. It
    requires the package `gmsh` to be installed.

    Examples:
        >>> import gmsh
        >>> import treams.io
        >>> gmsh.initialize()
        >>> gmsh.model.add("spheres")
        >>> treams.io.mesh_spheres([1, 2], [[0, 0, 2], [0, 0, -2]], gmsh.model)
        <class 'gmsh.model'>
        >>> gmsh.write("spheres.msh")  # doctest: +SKIP
        >>> gmsh.finalize()  # doctest: +SKIP

    Args:
        radii (float, array_like): Radii of the spheres.
        positions (float, (N, 3)-array): Positions of the spheres.
        model (gmsh.model): Gmsh model to modify.
        meshsize (float, optional): Mesh size, if None a fifth of the largest radius is
            used.
        meshsize (float, optional): Mesh size of the surfaces, if left empty it is set
            equal to the general mesh size.

    Returns:
        gmsh.model
    """
    if meshsize is None:
        meshsize = np.max(radii) * 0.2
    if meshsize_boundary is None:
        meshsize_boundary = meshsize

    spheres = []
    for i, (radius, position) in enumerate(zip(radii, positions)):
        tag = i + 1
        model.occ.addSphere(*position, radius, tag)
        spheres.append((3, tag))

    model.occ.synchronize()
    for _, tag in spheres:
        model.addPhysicalGroup(3, [tag], tag)
        # Add surfaces for other mesh formats like stl, ...
        model.addPhysicalGroup(2, [tag], tag)

    model.mesh.setSize(model.getEntities(0), meshsize)
    model.mesh.setSize(
        model.getBoundary(spheres, False, False, True), meshsize_boundary
    )
    return model


def _translate_polarizations(pols, poltype=None):
    """Translate the polarization index into words.

    The indices 0 and 1 are translated to "negative" and "positive", respectively, when
    helicity modes are chosen. For parity modes they are translated to "magnetic" and
    "electric".

    Args:
        pols (int, array_like): Array of indices 0 and 1.
        poltype (str, optional): Polarization type (:ref:`polarizations.Polarizations`).

    Returns:
        list[str]
    """
    poltype = treams.config.POLTYPE if poltype is None else poltype
    if poltype == "helicity":
        names = ["negative", "positive"]
    elif poltype == "parity":
        names = ["magnetic", "electric"]
    else:
        raise ValueError("unrecognized poltype")
    return [names[i] for i in pols]


def _translate_polarizations_inv(pols):
    """Translate the polarization into indices.

    This function is the inverse of :func:`treams.io._translate_polarizations`. The
    words "negative" and "minus" are translated to 0 and the words "positive" and "plus"
    are translated to 1, if helicity modes are chosen. For parity modes, modes
    "magnetic" or "te" are translated to 0 and the modes "electric" or "tm" to 1.

    Args:
        pols (string, array_like): Array of strings

    Returns:
        tuple[list[int], str]
    """
    helicity = {"plus": 1, "positive": 1, "minus": 0, "negative": 0}
    parity = {"te": 0, "magnetic": 0, "tm": 1, "electric": 1, "M": 0, "N": 1}
    if pols[0].decode() in helicity:
        dct = helicity
        poltype = "helicity"
    elif pols[0].decode() in parity:
        dct = parity
        poltype = "parity"
    else:
        raise ValueError(f"unrecognized polarization '{pols[0].decode()}'")
    return [dct[i.decode()] for i in pols], poltype


def _remove_leading_ones(shape):
    while len(shape) > 0 and shape[0] == 1:
        shape = shape[1:]
    return shape


def _collapse_dims(arr):
    for i in range(arr.ndim):
        test = arr[(slice(None),) * i + (slice(1),)]
        if np.all(arr == test):
            arr = test
    return arr.reshape(_remove_leading_ones(arr.shape))


def save_hdf5(
    h5file,
    tms,
    name="",
    description="",
    keywords="",
    embedding_group=None,
    embedding_name="",
    embedding_description="",
    embedding_keywords="",
    uuid=None,
    uuid_version=4,
    lunit="nm",
):
    """Save a set of T-matrices in a HDF5 file.

    With an open and writeable datafile, this function stores the main parts of as
    T-matrix in the file. It is left open for the user to add additional metadata.

    Args:
        h5file (h5py.Group): A HDF5 file opened with h5py.
        tms (TMatrix, array_like): Array of T-matrix instances.
        name (str): Name to add to the file as attribute.
        description (str): Description to add to file as attribute.
        keywords (str): Keywords to add to file as attribute.
        embedding_group (h5py.Group, optional): Group object for the embedding material,
            defaults to "/materials/embedding/".
        embedding_name (string, optional): Name of the embedding material.
        embedding_description (string, optional): Description of the embedding material.
        embedding_keywords (string, optional): Keywords for the embedding material.
        uuid (bytes, optional): UUID of the file, a new one is created if omitted.
        uuid_version (int, optional): UUID version.
        lunit (string, optional): Length unit used for the positions and (as
            inverse) for the wave number.
    """
    tms_arr = np.array(tms)
    if tms_arr.dtype == object:
        raise ValueError("can only save T-matrices of the same size")
    tms_obj = np.empty(tms_arr.shape[:-2], object)
    tms_obj[:] = tms
    tm = tms_obj.flat[0]
    basis = tm.basis
    poltype = tm.poltype

    k0s = np.zeros((tms_arr.shape)[:-2])
    epsilon = np.zeros((tms_arr.shape)[:-2], complex)
    mu = np.zeros((tms_arr.shape)[:-2], complex)
    if poltype == "helicity":
        kappa = np.zeros((tms_arr.shape)[:-2], complex)

    for i, tm in enumerate(tms_obj.flat):
        if poltype != tm.poltype or basis != tm.basis:
            raise ValueError(
                "incompatible T-matrices: mixed poltypes or different bases"
            )
        k0s.flat[i] = tm.k0
        epsilon.flat[i] = tm.material.epsilon
        mu.flat[i] = tm.material.mu
        if poltype == "helicity":
            kappa.flat[i] = tm.material.kappa

    h5file["tmatrix"] = tms_arr
    if uuid is None:
        h5file["uuid"] = np.void(_uuid.uuid4().bytes)
        h5file["uuid"].attrs["version"] = 4
    else:
        h5file["uuid"] = uuid
        h5file["uuid"].attrs["version"] = uuid_version

    _name_descr_kw(h5file, name, description, keywords)
    h5file["angular_vacuum_wavenumber"] = _collapse_dims(k0s)
    h5file["angular_vacuum_wavenumber"].attrs["unit"] = lunit + r"^{-1}"
    h5file["modes/l"] = basis.l
    h5file["modes/m"] = basis.m
    h5file["modes/polarization"] = _translate_polarizations(basis.pol, poltype)
    if any(basis.pidx != 0):
        h5file["modes/pidx"] = basis.pidx
    if not np.array_equiv(basis.positions, [[0, 0, 0]]):
        h5file["modes/positions"] = basis.positions
        h5file["modes/positions"].attrs["unit"] = lunit
    if embedding_group is None:
        embedding_group = h5file.create_group("materials/embedding")
    embedding_group["relative_permittivity"] = _collapse_dims(epsilon)
    embedding_group["relative_permeability"] = _collapse_dims(mu)
    if poltype == "helicity":
        embedding_group["chirality"] = _collapse_dims(kappa)
    _name_descr_kw(
        embedding_group, embedding_name, embedding_description, embedding_keywords
    )
    h5file["embedding"] = h5py.SoftLink(embedding_group.name)

    h5file.attrs["created_with"] = (
        f"python={sys.version.split()[0]},"
        f"h5py={version('h5py')},"
        f"treams={version('treams')}"
    )
    h5file.attrs["storage_format_version"] = "0.0.1-4-g1266244"


def _name_descr_kw(fobj, name, description="", keywords=""):
    for key, val in [
        ("name", name),
        ("description", description),
        ("keywords", keywords),
    ]:
        val = str(val)
        if val != "":
            fobj.attrs[key] = val


def _convert_to_k0(x, xtype, xunit, k0unit=r"nm^{-1}"):
    c = 299792458.0
    k0unit = INVLENGTHS[k0unit]
    if xtype == "frequency":
        xunit = FREQUENCIES[xunit]
        return 2 * np.pi * x / c * (xunit / k0unit)
    if xtype == "angular_frequency":
        xunit = FREQUENCIES[xunit]
        return x / c * (xunit / k0unit)
    if xtype == "angular_vacuum_wavelength":
        xunit = INVLENGTHS[xunit]
        return x * (xunit / k0unit)
    if xtype == "angular_vacuum_wavenumber":
        xunit = LENGTHS[xunit]
        return 2 * np.pi / (x * xunit * k0unit)
    raise ValueError(f"unrecognized frequency/wavenumber/wavelength type: {xtype}")


def load_hdf5(filename, lunit="nm"):
    """Load a T-matrix stored in a HDF4 file.

    Args:
        filename (str or h5py.Group): Name of the h5py file or a handle to a h5py group.
        lunit (str, optional): Unit of length to be used in the T-matrices.

    Returns:
        np.ndarray[TMatrix]
    """
    if isinstance(filename, h5py.Group):
        return _load_hdf5(filename, lunit)
    with h5py.File(filename, "r") as f:
        return _load_hdf5(f, lunit)


def _load_hdf5(h5file, lunit=None):
    for freq_type in (
        "frequency",
        "angular_frequency",
        "vacuum_wavelength",
        "vacuum_wavenumber",
        "angular_vacuum_wavenumber",
    ):
        if freq_type in h5file:
            ld_freq = h5file[freq_type][()]
            break
    else:
        raise ValueError("no definition of frequency found")
    if "modes/positions" in h5file:
        k0unit = h5file["modes/positions"].attrs.get("unit", lunit) + r"^{-1}"
    else:
        k0unit = lunit + r"^{-1}"
    tms = h5file["tmatrix"][...]
    k0s = _convert_to_k0(ld_freq, freq_type, h5file[freq_type].attrs["unit"], k0unit)

    epsilon = h5file.get("embedding/relative_permittivity", np.array(None))[()]
    mu = h5file.get("embedding/relative_permeability", np.array(None))[()]
    if epsilon is None is mu:
        n = h5file.get("embedding/refractive_index", np.array(1))[...]
        z = h5file.get("embedding/relative_impedance", 1 / n)[...]
        epsilon = n / z
        mu = n * z
    epsilon = 1 if epsilon is None else epsilon
    mu = 1 if mu is None else mu

    kappa = z = h5file.get("embedding/chirality_parameter", np.array(0))[...]

    positions = h5file.get("modes/positions", np.zeros((1, 3)))[...]

    l_inc = h5file.get("modes/l", np.array(None))[...]
    l_inc = h5file.get("modes/l_incident", l_inc)[()]
    m_inc = h5file.get("modes/m", np.array(None))[...]
    m_inc = h5file.get("modes/m_incident", m_inc)[()]
    pol_inc = h5file.get("modes/polarization", np.array(None))[...]
    pol_inc = h5file.get("modes/polarization_incident", pol_inc)[()]
    l_sca = h5file.get("modes/l", np.array(None))[...]
    l_sca = h5file.get("modes/l_scattered", l_sca)[()]
    m_sca = h5file.get("modes/m", np.array(None))[...]
    m_sca = h5file.get("modes/m_scattered", m_sca)[()]
    pol_sca = h5file.get("modes/polarization", np.array(None))[...]
    pol_sca = h5file.get("modes/polarization_scattered", pol_sca)[()]

    if any(x is None for x in (l_inc, l_sca, m_inc, m_sca, pol_inc, pol_sca)):
        raise ValueError("mode definition missing")

    pol_inc, poltype = _translate_polarizations_inv(pol_inc)
    pol_sca, poltype_sca = _translate_polarizations_inv(pol_sca)

    if poltype_sca != poltype:
        raise ValueError("different modetypes")

    pidx_inc = h5file.get("modes/position_index", np.zeros_like(l_inc))[()]
    pidx_inc = h5file.get("modes/positions_index_scattered", pidx_inc)[()]
    pidx_sca = h5file.get("modes/position_index", np.zeros_like(l_sca))[()]
    pidx_sca = h5file.get("modes/positions_index_scattered", pidx_sca)[()]

    shape = tms.shape[:-2]
    k0s = np.broadcast_to(k0s, shape)
    epsilon = np.broadcast_to(epsilon, shape)
    mu = np.broadcast_to(mu, shape)
    kappa = np.broadcast_to(kappa, shape)

    basis_inc = treams.SphericalWaveBasis(
        zip(pidx_inc, l_inc, m_inc, pol_inc), positions
    )
    basis_sca = treams.SphericalWaveBasis(
        zip(pidx_sca, l_sca, m_sca, pol_sca), positions
    )
    basis = basis_inc | basis_sca

    ix_inc = [basis.index(b) for b in basis_inc]
    ix_sca = [[basis.index(b)] for b in basis_sca]

    res = np.empty(shape, object)
    for i in np.ndindex(*shape):
        res[i] = treams.TMatrix(
            np.zeros((len(basis),) * 2, complex),
            k0=k0s[i],
            basis=basis,
            poltype=poltype,
            material=treams.Material(epsilon[i], mu[i], kappa[i]),
        )
        res[i][ix_sca, ix_inc] = tms[i]
    if not res.shape:
        res = res.item()
    return res



ctypedef fused real_t:
    double
    long

ctypedef fused number_t:
    double complex
    double

cdef real_t area(real_t *a, real_t *b) nogil
cdef long cube_next(long *r, long d, long n) nogil
cdef long cubeedge_next(long *r, long d, long n) nogil
cpdef diffr_orders_circle(real_t[:, :] b, double rmax)
cpdef long ncube(long d, long n) nogil
cpdef long nedge(long d, long n) nogil
cdef void recvec2(double *a0, double *a1, double *b0, double *b1) nogil
cdef void recvec3(double *a0, double *a1, double *a2, double *b0, double *b1, double *b2) nogil
cdef real_t volume(real_t *a, real_t *b, real_t *c) nogil

cpdef double complex lsumcw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex lsumcw1d_shift(long l, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex lsumcw2d(long l, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cpdef double complex lsumsw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex lsumsw1d_shift(long l, long m, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex lsumsw2d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex lsumsw2d_shift(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex lsumsw3d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil

cpdef double complex realsumcw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex realsumcw1d_shift(long l, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex realsumcw2d(long l, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cpdef double complex realsumsw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex realsumsw1d_shift(long l, long m, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex realsumsw2d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex realsumsw2d_shift(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex realsumsw3d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil

cpdef double complex recsumcw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex recsumcw1d_shift(long l, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex recsumcw2d(long l, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cpdef double complex recsumsw1d(long l, number_t k, double kpar, double a, double r, number_t eta) nogil
cdef double complex recsumsw1d_shift(long l, long m, number_t k, double kpar, double a, double *r, number_t eta) nogil
cdef double complex recsumsw2d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex recsumsw2d_shift(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil
cdef double complex recsumsw3d(long l, long m, number_t k, double *kpar, double *a, double *r, number_t eta) nogil

cpdef double complex zero3d(double complex eta) nogil
cpdef double complex zero2d(double complex eta) nogil

cpdef double complex dsumcw1d(long l, number_t k, double kpar, double a, double r, long i) nogil
cdef double complex dsumcw1d_shift(long l, number_t k, double kpar, double a, double *r, long i) nogil
cdef double complex dsumcw2d(long l, number_t k, double *kpar, double *a, double *r, long i) nogil
cpdef double complex dsumsw1d(long l, number_t k, double kpar, double a, double r, long i) nogil
cdef double complex dsumsw1d_shift(long l, long m, number_t k, double kpar, double a, double *r, long i) nogil
cdef double complex dsumsw2d(long l, long m, number_t k, double *kpar, double *a, double *r, long i) nogil
cdef double complex dsumsw2d_shift(long l, long m, number_t k, double *kpar, double *a, double *r, long i) nogil
cdef double complex dsumsw3d(long l, long m, number_t k, double *kpar, double *a, double *r, long i) nogil



import matplotlib.pyplot as plt
import numpy as np

import treams

k0 = 2 * np.pi / 1000
materials = [treams.Material(16 + 0.5j), treams.Material()]
lmax = 3
radii = [110, 90, 80, 75]
positions = (220 / np.sqrt(24)) * np.array(
    [
        [np.sqrt(8), 0, -1],
        [-np.sqrt(2), np.sqrt(6), -1],
        [-np.sqrt(2), -np.sqrt(6), -1],
        [0, 0, 3],
    ]
)

spheres = [treams.TMatrix.sphere(lmax, k0, r, materials) for r in radii]
tm = treams.TMatrix.cluster(spheres, positions).interaction.solve()

inc = treams.plane_wave([k0, 0, 0], 0, k0=tm.k0, material=tm.material)
sca = tm @ inc.expand(tm.basis)
xs = tm.xs(inc)

grid = np.mgrid[-300:300:101j, 0:1, -300:300:101j].squeeze().transpose((1, 2, 0))
intensity = np.zeros_like(grid[..., 0])
valid = tm.valid_points(grid, radii)
intensity[~valid] = np.nan
intensity[valid] = 0.5 * np.sum(
    np.abs(inc.efield(grid[valid]) + sca.efield(grid[valid])) ** 2, -1
)

fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    grid[0, :, 2], grid[:, 0, 0], intensity.T, shading="nearest", vmin=0, vmax=2,
)
cb = plt.colorbar(pcm)
cb.set_label("Intensity")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
ax.text(
    0,
    0,
    r"$\sigma_\mathrm{sca} = "
    f"{xs[0]:.7}\\,$nm$^2$\n"
    r"$\sigma_\mathrm{ext} = "
    f"{xs[1]:.7}\\,$nm$^2$",
)
fig.show()

tm_global = tm.expand(treams.SphericalWaveBasis.default(6))
sca = tm_global @ inc.expand(tm_global.basis)

xs = tm_global.xs(inc)

grid = np.mgrid[-300:300:101j, 0:1, -300:300:101j].squeeze().transpose((1, 2, 0))
intensity_global = np.zeros_like(grid[..., 0])
valid = tm_global.valid_points(grid, [260])
intensity_global[~valid] = np.nan
intensity_global[valid] = 0.5 * np.sum(
    np.abs(inc.efield(grid[valid]) + sca.efield(grid[valid])) ** 2, -1
)

fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    grid[0, :, 2], grid[:, 0, 0], intensity_global.T, shading="nearest", vmin=0, vmax=2,
)
cb = plt.colorbar(pcm)
cb.set_label("Intensity")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
ax.text(
    0,
    0,
    r"$\sigma_\mathrm{sca} = "
    f"{xs[0]:.7}\\,$nm$^2$\n"
    r"$\sigma_\mathrm{ext} = "
    f"{xs[1]:.7}\\,$nm$^2$",
)
fig.show()

inc = treams.plane_wave([0, 0, -k0], 0, k0=tm.k0, material=tm.material)
tm_rotate = tm_global.rotate(0, np.pi / 2)
sca = tm_rotate @ inc.expand(tm_rotate.basis)

xs = tm_rotate.xs(inc)

grid = np.mgrid[-300:300:101j, 0:1, -300:300:101j].squeeze().transpose((1, 2, 0))
intensity_global = np.zeros_like(grid[..., 0])
valid = tm_rotate.valid_points(grid, [260])
intensity_global[~valid] = np.nan
intensity_global[valid] = 0.5 * np.sum(
    np.abs(inc.efield(grid[valid]) + sca.efield(grid[valid])) ** 2, -1
)

fig, ax = plt.subplots()
pcm = ax.pcolormesh(
    grid[0, :, 2], grid[:, 0, 0], intensity_global.T, shading="nearest", vmin=0, vmax=2,
)
cb = plt.colorbar(pcm)
cb.set_label("Intensity")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
ax.text(
    0,
    0,
    r"$\sigma_\mathrm{sca} = "
    f"{xs[0]:.7}\\,$nm$^2$\n"
    r"$\sigma_\mathrm{ext} = "
    f"{xs[1]:.7}\\,$nm$^2$",
)
fig.show()



import numpy as np
import scipy.special as sc

from treams import cw


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class TestPeriodicToPw:
    def test(self):
        assert isclose(
            cw.periodic_to_pw(6, -5, 4, 1, 4, 3, 1, 2),
            2 * np.power((-5 - 6j) / np.sqrt(61), 3) / 10,
        )

    def test_ky_zero(self):
        assert isclose(cw.periodic_to_pw(6, 0, 4, 1, 4, 3, 1, 2), 5e19 + 5e19j)

    def test_pol(self):
        assert cw.periodic_to_pw(6, 5j, 4, 1, 4, 3, 0, 2) == 0


class TestRotate:
    def test(self):
        assert isclose(cw.rotate(3, 2, 1, 3, 2, 1, 4), np.exp(8j))

    def test_zero(self):
        assert cw.rotate(3, 2, 1, 2, 2, 1, 4) == 0


class TestToSw:
    def test_h(self):
        assert isclose(cw.to_sw(4, 3, 1, 2, 3, 1, 5), -2.481668635232552j)

    def test_h_zero(self):
        assert cw.to_sw(4, 3, 1, 2, 2, 1, 5) == 0

    def test_h_pol(self):
        assert cw.to_sw(4, 3, 1, 2, 3, 0, 5) == 0

    def test_p_same(self):
        assert isclose(
            cw.to_sw(4, 3, 1, 2, 3, 1, 5, poltype="parity"), 1.063572272242522j
        )

    def test_p_same_zero(self):
        assert cw.to_sw(4, 3, 1, 2, 2, 1, 5, poltype="parity") == 0

    def test_p_opposite(self):
        assert isclose(
            cw.to_sw(4, 3, 1, 2, 3, 0, 5, poltype="parity"), -3.545240907475074j
        )

    def test_p_opposite_zero(self):
        assert cw.to_sw(4, 3, 1, 2, 2, 0, 5, poltype="parity") == 0


class TestTranslate:
    def test_s(self):
        assert isclose(
            cw.translate(3, 2, 1, 3, -2, 1, 4, 5, 6), sc.hankel1(-4, 4) * np.exp(-2j)
        )

    def test_s_opposite(self):
        assert cw.translate(3, 2, 0, 0, 2, 1, 4 + 1j, 5, 6) == 0

    def test_s_zero(self):
        assert cw.translate(3, 2, 0, 0, 2, 1, 0, 5, 0) == 0

    def test_r(self):
        assert isclose(
            cw.translate(3, 2, 1, 3, -2, 1, 4, 5, 6, singular=False),
            sc.jv(-4, 4) * np.exp(-2j),
        )

    def test_r_opposite(self):
        assert cw.translate(3, 2, 0, 0, 2, 1, 4 + 1j, 5, 6, singular=False) == 0


class TestTranslatePeriodic:
    def test(self):
        assert isclose(
            cw.translate_periodic(1, 0, 2, [0, 0, 0], ([0], [0], [0]))[0, 0],
            0.76102358j,
        )

    def test_ks(self):
        assert isclose(
            cw.translate_periodic([1, 2], 0, 2, [0, 0, 0], ([1], [3], [1]))[0, 0],
            -0.42264973081037416 + 0.30599871466756257j,
        )

    def test_ks_same(self):
        assert isclose(
            cw.translate_periodic(
                [1, 1], [0, 1], [[2, 0], [0, 2]], [0, 0, 0], ([0], [0], [0])
            )[0, 0],
            -1.0 + 0.28364898j,
        )



import matplotlib.pyplot as plt
import numpy as np

import treams

k0s = 2 * np.pi * np.linspace(1 / 600, 1 / 350, 100)
material_slab = 3
thickness = 10
material_sphere = (4, 1, 0.05)
pitch = 500
lattice = treams.Lattice.square(pitch)
radius = 100
lmax = mmax = 3

tr = np.zeros((len(k0s), 2))
for i, k0 in enumerate(k0s):
    kpar = [0, 0.3 * k0]

    spheres = treams.TMatrix.sphere(
        lmax, k0, radius, [material_sphere, 1]
    ).latticeinteraction.solve(pitch, kpar[0])
    cwb = treams.CylindricalWaveBasis.diffr_orders(kpar[0], mmax, pitch, 0.02)
    spheres_cw = treams.TMatrixC.from_array(spheres, cwb)
    chain_cw = spheres_cw.latticeinteraction.solve(pitch, kpar[1])

    pwb = treams.PlaneWaveBasisByComp.diffr_orders(kpar, lattice, 0.02)
    plw = treams.plane_wave(kpar, [1, 0, 0], k0=k0, basis=pwb, material=1)
    slab = treams.SMatrices.slab(thickness, pwb, k0, [1, material_slab, 1])
    dist = treams.SMatrices.propagation([0, 0, radius], pwb, k0, 1)
    array = treams.SMatrices.from_array(chain_cw, pwb)
    total = treams.SMatrices.stack([slab, dist, array])
    tr[i, :] = total.tr(plw)

fig, ax = plt.subplots()
ax.set_xlabel("Frequency (THz)")
ax.plot(299792.458 * k0s / (2 * np.pi), tr[:, 0])
ax.plot(299792.458 * k0s / (2 * np.pi), tr[:, 1])
ax.legend(["$T$", "$R$"])
fig.show()



import numpy as np

from treams import misc


class TestRefractiveIndex:
    def test_single(self):
        assert np.all(misc.refractive_index() == [1, 1])

    def test_multiple(self):
        expect = [
            [[0.5, 1.5], [-np.sqrt(2) + 1j, np.sqrt(2) + 1j]],
            [[np.sqrt(3) - 0.5, np.sqrt(3) + 0.5], [1j, 3j]],
        ]
        assert np.all(
            misc.refractive_index(mu=[[1, 2], [3, -4 + 0j]], kappa=[0.5, 1j]) == expect
        )


class TestBasisChange:
    def test(self):
        assert np.all(
            misc.basischange(([1, 1], [0, 1]))
            == np.sqrt(0.5) * np.array([[-1, 1], [1, 1]])
        )


class TestPickModes:
    def test(self):
        assert np.all(misc.pickmodes(([1, 1], [0, 1]), ([1], [0])) == [[1], [0]])


class TestWaveVecZ:
    def test(self):
        assert misc.wave_vec_z(3, 4, 5) == 0

    def test_neg_single(self):
        assert misc.wave_vec_z(0, 0, 2 - 2j) == -2 + 2j

    def test_multiple(self):
        assert np.all(misc.wave_vec_z([3, 3], [0, 5], 5) == [4, 3j])


class TestFirstBrillouin1d:
    def test(self):
        assert np.abs(misc.firstbrillouin1d(0.81, 0.5) + 0.19) < 1e-16

    def test_edge(self):
        assert misc.firstbrillouin1d(-0.25, 0.5) == 0.25


class TestFirstBrillouin2d:
    def test(self):
        assert np.all(
            np.abs(
                misc.firstbrillouin2d([1.81, -0.25], 0.5 * np.eye(2)) - [-0.19, -0.25]
            )
            < 1e-16
        )


class TestFirstBrillouin3d:
    def test(self):
        assert np.all(
            np.abs(
                misc.firstbrillouin3d([1.81, -0.25, 0], 0.5 * np.eye(3))
                - [-0.19, -0.25, 0]
            )
            < 1e-16
        )



import numpy as np
import pytest
import scipy.special as ssc

import treams.special as sc
import treams.special.cython_special as cs


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


EPS = 2e-7
EPSSQ = 4e-14


class TestIncgamma:
    def test_zero_real(self):
        assert isclose(sc.incgamma(0, 1.5), 0.10001958240663263, rel_tol=EPSSQ)

    def test_zero_complex(self):
        assert isclose(
            sc.incgamma(0, 2 + 4j),
            0.006575211740584215 + 0.0261438237000811j,
            rel_tol=EPSSQ,
        )

    def test_zero_negreal(self):
        assert isclose(
            sc.incgamma(0, -3 + 0.0j),
            -9.933832570625414 - 3.141592653589793j,
            rel_tol=EPSSQ,
        )

    def test_zero_negreal_branch(self):
        assert isclose(
            sc.incgamma(0, complex(-3, -0.0)),
            -9.933832570625414 + 3.141592653589793j,
            rel_tol=EPSSQ,
        )

    def test_half_real(self):
        assert isclose(sc.incgamma(0.5, 1.5), 0.1475825132040964, rel_tol=EPSSQ)

    def test_half_complex(self):
        assert isclose(
            sc.incgamma(0.5, 2 + 4j),
            -0.01415763494202471 + 0.058731665238669344j,
            rel_tol=EPSSQ,
        )

    def test_half_negreal(self):
        assert isclose(
            sc.incgamma(0.5, -3 + 0.0j),
            1.7724538509055152 - 14.626171384019093j,
            rel_tol=EPSSQ,
        )

    def test_half_negreal_branch(self):
        assert isclose(
            sc.incgamma(0.5, complex(-3, -0.0)),
            1.7724538509055152 + 14.626171384019093j,
            rel_tol=EPSSQ,
        )

    def test_one_real(self):
        assert isclose(sc.incgamma(1, 1.5), 0.22313016014842982, rel_tol=EPSSQ)

    def test_one_complex(self):
        assert isclose(
            sc.incgamma(1, 2 + 4j),
            -0.08846104456538201 + 0.10242208005667372j,
            rel_tol=EPSSQ,
        )

    def test_one_negreal(self):
        assert isclose(sc.incgamma(1, -3), 20.085536923187668, rel_tol=EPSSQ)

    def test_one_negreal_branch(self):
        assert isclose(
            sc.incgamma(1, complex(-3, -0.0)), 20.085536923187668, rel_tol=EPSSQ
        )

    def test_ten_real(self):
        assert isclose(sc.incgamma(10, 1.5), 362878.5130988457, rel_tol=EPSSQ)

    def test_ten_complex(self):
        assert isclose(
            sc.incgamma(10, 2 + 4j),
            345166.2033113096 - 45890.997544067905j,
            rel_tol=EPSSQ,
        )

    def test_ten_negreal(self):
        assert isclose(sc.incgamma(10, -3), 270070.1294691813, rel_tol=EPSSQ)

    def test_ten_negreal_branch(self):
        assert isclose(
            sc.incgamma(10, complex(-3, -0.0)), 270070.1294691813, rel_tol=EPSSQ
        )

    def test_neg_real(self):
        assert isclose(sc.incgamma(-10, 1.5), 0.0003324561166899859, rel_tol=EPSSQ)

    def test_neg_complex(self):
        assert isclose(
            sc.incgamma(-10, 2 + 4j),
            -3.109457703343637e-9 - 9.73849356067146e-10j,
            rel_tol=EPSSQ,
        )

    def test_neg_negreal(self):
        assert isclose(
            sc.incgamma(-10, -3 + 0.0j),
            0.00005342780082756921 - 8.657387162670285e-7j,
            rel_tol=EPSSQ,
        )

    def test_neg_negreal_branch(self):
        assert isclose(
            sc.incgamma(-10, complex(-3, -0.0)),
            0.00005342780082756921 + 8.657387162670285e-7j,
            rel_tol=EPSSQ,
        )

    def test_huge(self):
        assert isclose(
            sc.incgamma(100, 120 - 42j),
            -4.4658781836612545e156 + 1.1325309029755172e155j,
            rel_tol=EPS,
        )

    def test_tiny(self):
        assert isclose(
            sc.incgamma(100, 1e-12 + 1e-3j),
            9.332621544394404e155 + 9.901988480274498e-306j,
            rel_tol=EPSSQ,
        )

    def test_huge_neg(self):
        assert isclose(
            sc.incgamma(-100, 42 - 12j),
            -2.2632633469219943e-185 + 3.010799952817908e-185j,
            rel_tol=EPS,
        )

    def test_tiny_neg(self):
        assert isclose(
            sc.incgamma(-100, 1e-2 + 1e-2j),
            -8.792072241883211e182 + 8.881173950746141e180j,
            rel_tol=EPS,
        )

    def test_huge_half(self):
        assert isclose(
            sc.incgamma(99.5, 120 - 42j),
            -3.9035577320280324e155 - 5.255135904002826e154j,
            rel_tol=EPSSQ,
        )

    def test_tiny_half(self):
        assert isclose(
            sc.incgamma(99.5, 1e-8 + 1e-3j),
            9.367802114655592e154 + 2.249528564769392e-301j,
            rel_tol=EPS,
        )

    def test_huge_neg_half(self):
        assert isclose(
            sc.incgamma(-99.5, 42 - 12j),
            -1.2102237586206635e-184 + 2.1853661691833773e-184j,
            rel_tol=EPS,
        )

    def test_tiny_neg_half(self):
        assert isclose(
            sc.incgamma(-99.5, 1e-2 + 1e-2j),
            -9.748868790182211e181 - 3.9232159891109815e181j,
            rel_tol=EPS,
        )

    @pytest.mark.skip
    def test_error(self):
        with pytest.raises(ValueError):
            sc.incgamma(0.4, 1)

    def test_zero_zero(self):
        result = sc.incgamma(0, 0)
        assert np.isinf(result.real) and result.imag == 0

    def test_neg_zero(self):
        result = sc.incgamma(-1, 0)
        assert np.isinf(result.real) and result.imag == 0

    def test_gamma(self):
        assert isclose(sc.incgamma(4, 0), 6)


class TestWigner3j:
    def test_zeros(self):
        assert isclose(sc.wigner3j(0, 0, 0, 0, 0, 0), 1)

    def test_small(self):
        assert isclose(sc.wigner3j(1, 0, 1, 0, 0, 0), -1 / np.sqrt(3))

    def test_medium(self):
        assert isclose(sc.wigner3j(5, 2, 6, 3, -2, -1), np.sqrt(8 / 1001))

    def test_large(self):
        assert isclose(sc.wigner3j(123, 60, 95, -64, 32, 32), -0.007933910368778899)

    def test_large_forward(self):
        assert isclose(sc.wigner3j(123, 60, 66, -64, 32, 32), -0.01069952664596778)

    def test_more(self):
        assert isclose(sc.wigner3j(43, 61, 21, -1, 12, -11), -0.02786405482897469)

    def test_even_more(self):
        assert isclose(sc.wigner3j(43, 61, 22, -9, -12, 21), -0.00005702392146902575)

    @pytest.mark.skip
    def test_error(self):
        with pytest.raises(ValueError):
            sc.wigner3j(-2, 1, 3, 0, 0, 0)

    def test_triangular(self):
        assert sc.wigner3j(3, 1, 1, 0, 0, 0) == 0.0

    def test_physical_jm(self):
        assert sc.wigner3j(1, 1, 1, 0, -2, 2) == 0.0

    def test_physical_m(self):
        assert sc.wigner3j(1, 1, 1, 0, 0, 1) == 0.0

    def test_zero_return(self):
        assert sc.wigner3j(2, 2, 1, 0, 0, 0) == 0.0

    def test_divide_by_zero(self):
        assert isclose(sc.wigner3j(4, 4, 1, 1, -1, 0), -1 / (6 * np.sqrt(5)))

    def test_last_init(self):
        assert isclose(sc.wigner3j(4, 7, 4, 1, 3, -4), np.sqrt(5 / 858))


class TestHankel1:
    def test(self):
        assert sc.hankel1(1, 3 + 1j) == ssc.hankel1(1, 3 + 1j)


class TestHankel2:
    def test(self):
        assert sc.hankel2(1, 3 + 1j) == ssc.hankel2(1, 3 + 1j)


class TestJv:
    def test_real_type(self):
        assert isinstance(sc.jv(1, 3), float)

    def test_real(self):
        assert sc.jv(1, 3) == ssc.jv(1, 3)

    def test_complex_type(self):
        assert isinstance(sc.jv(1, 3 + 0j), complex)

    def test_complex(self):
        assert sc.jv(1, 3 + 1j) == ssc.jv(1, 3 + 1j)


class TestSpericalJn:
    def test_real_type(self):
        assert isinstance(sc.spherical_jn(1, 3), float)

    def test_real(self):
        assert sc.spherical_jn(1, 3) == ssc.spherical_jn(1, 3)

    def test_complex_type(self):
        assert isinstance(sc.spherical_jn(1, 3 + 0j), complex)

    def test_complex(self):
        assert sc.spherical_jn(1, 3 + 1j) == ssc.spherical_jn(1, 3 + 1j)


class TestSpericalYn:
    def test_real_type(self):
        assert isinstance(sc.spherical_yn(1, 3), float)

    def test_real(self):
        assert sc.spherical_yn(1, 3) == ssc.spherical_yn(1, 3)

    def test_complex_type(self):
        assert isinstance(sc.spherical_yn(1, 3 + 0j), complex)

    def test_complex(self):
        assert sc.spherical_yn(1, 3 + 1j) == ssc.spherical_yn(1, 3 + 1j)


class TestYv:
    def test_real_type(self):
        assert isinstance(sc.yv(1, 3), float)

    def test_real(self):
        assert sc.yv(1, 3) == ssc.yv(1, 3)

    def test_complex_type(self):
        assert isinstance(sc.yv(1, 3 + 0j), complex)

    def test_complex(self):
        assert sc.yv(1, 3 + 1j) == ssc.yv(1, 3 + 1j)


class TestLpmv:
    def test_real_type(self):
        assert isinstance(sc.lpmv(1, 2, 0.3), float)

    def test_real(self):
        assert sc.lpmv(1, 2, 0.3) == ssc.lpmv(1, 2, 0.3)

    def test_complex_type(self):
        assert isinstance(sc.lpmv(1, 2, 0.3 + 0j), complex)

    @pytest.mark.skip
    def test_complex_error(self):
        with pytest.raises(ValueError):
            sc.lpmv(0, -1, 0j)

    def test_non_physical(self):
        assert sc.lpmv(2, 1, 0j) == 0

    def test_complex_00(self):
        assert sc.lpmv(0, 0, 4j) == 1

    def test_complex_01(self):
        assert sc.lpmv(0, 1, 3 + 0j) == 3

    def test_complex_even(self):
        assert isclose(
            sc.lpmv(0, 14, 1 / np.sqrt(2) + 0j), ssc.lpmv(0, 14, 1 / np.sqrt(2))
        )

    def test_complex_odd(self):
        assert isclose(
            sc.lpmv(0, 13, -1.2 + 0j), ssc.clpmn(0, 13, -1.2, type=2)[0][0, 13]
        )

    def test_complex_odd_c(self):
        assert isclose(
            sc.lpmv(0, 13, -1.2 - 1j), ssc.clpmn(0, 13, -1.2 - 1j, type=2)[0][0, 13]
        )

    def test_complex_asso(self):
        assert isclose(
            sc.lpmv(3, 3, 0.1 + 0j), ssc.clpmn(3, 3, -0.1 + 0j, type=2)[0][3, 3]
        )

    def test_complex_asso_above(self):
        assert isclose(
            sc.lpmv(-5, 12, -1.3 + 0j),
            ssc.clpmn(-5, 12, -1.3 + 1e-16j, type=2)[0][5, 12],
        )

    def test_complex_asso_below(self):
        assert isclose(
            sc.lpmv(-5, 12, complex(-1.3, -0.0)),
            ssc.clpmn(-5, 12, complex(-1.3, 0), type=2)[0][5, 12],
        )

    def test_complex_asso_odd(self):
        assert isclose(
            sc.lpmv(-4, 13, 1.3 - 3j), ssc.clpmn(-4, 13, 1.3 - 3j, type=2)[0][4, 13]
        )

    def test_complex_asso_pos(self):
        assert isclose(
            sc.lpmv(5, 12, 0.4 + 0j), ssc.clpmn(5, 12, 0.4 + 0j, type=2)[0][5, 12]
        )

    def test_complex_asso_pos_c(self):
        assert isclose(
            sc.lpmv(5, 12, 1.3 - 1j), ssc.clpmn(5, 12, 1.3 - 1j, type=2)[0][5, 12]
        )

    def test_nan_l(self):
        assert np.isnan(sc.lpmv(0, 0.5, 1j))

    def test_nan_m(self):
        assert np.isnan(sc.lpmv(0.5, 1, 1j))


class TestSphHarm:
    def test_real(self):
        assert sc.sph_harm(1, 2, 3, 4) == ssc.sph_harm(1, 2, 3, 4)

    def test_complex(self):
        assert isclose(
            sc.sph_harm(1, 2, 3, 4j),
            np.sqrt(5 / (24 * np.pi))
            * ssc.clpmn(1, 2, np.cos(4j), type=2)[0][1, 2]
            * np.exp(3j),
        )

    def test_nan_l(self):
        assert np.isnan(sc.sph_harm(0, 0.5, 0, 1j))

    def test_nan_m(self):
        assert np.isnan(sc.sph_harm(0.5, 1, 0, 1j))


class TestSphHankel1:
    def test(self):
        assert isclose(
            sc.spherical_hankel1(1, 2 + 1j), 0.0589704984384257 - 0.1995739736279250j
        )


class TestSphHankel2:
    def test(self):
        assert isclose(
            sc.spherical_hankel2(1, 2 + 1j), 1.0624415871431773 + 0.2312289984789762j
        )


class TestJvD:
    def test_zero(self):
        assert isclose(sc.jv_d(0, 3), ssc.jvp(0, 3))

    def test_non_zero(self):
        assert isclose(sc.jv_d(2, 3 + 4j), ssc.jvp(2, 3 + 4j))


class TestYvD:
    def test_zero(self):
        assert isclose(sc.yv_d(0, 3), ssc.yvp(0, 3))

    def test_non_zero(self):
        assert isclose(sc.yv_d(2, 3 + 4j), ssc.yvp(2, 3 + 4j))


class TestHankel1D:
    def test_zero(self):
        assert isclose(sc.hankel1_d(0, 3), ssc.h1vp(0, 3))

    def test_non_zero(self):
        assert isclose(sc.hankel1_d(2, 3 + 4j), ssc.h1vp(2, 3 + 4j))


class TestHankel2D:
    def test_zero(self):
        assert isclose(sc.hankel2_d(0, 3), ssc.h2vp(0, 3))

    def test_non_zero(self):
        assert isclose(sc.hankel2_d(2, 3 + 4j), ssc.h2vp(2, 3 + 4j))


class TestSphericalJnD:
    def test_zero(self):
        assert isclose(sc.spherical_jn_d(0, 3), ssc.spherical_jn(0, 3, True))

    def test_non_zero(self):
        assert isclose(sc.spherical_jn_d(2, 3 + 4j), ssc.spherical_jn(2, 3 + 4j, True))

    def test_zero_arg(self):
        assert isclose(sc.spherical_jn_d(3, 0), ssc.spherical_jn(3, 0, True))

    def test_zero_arg_2(self):
        assert isclose(sc.spherical_jn_d(1, 0), ssc.spherical_jn(1, 0, True))


class TestSphericalYnD:
    def test_zero(self):
        assert isclose(sc.spherical_yn_d(0, 3), ssc.spherical_yn(0, 3, True))

    def test_non_zero(self):
        assert isclose(sc.spherical_yn_d(2, 3 + 4j), ssc.spherical_yn(2, 3 + 4j, True))


class TestSphericalHankel1D:
    def test_zero(self):
        assert isclose(
            sc.spherical_hankel1_d(0, 3), -0.3456774997623560 - 0.0629591636023160j
        )

    def test_non_zero(self):
        assert isclose(
            sc.spherical_hankel1_d(2, 3 + 4j),
            0.005890898262660627 - 0.003935166138594885j,
        )


class TestSphericalHankel2D:
    def test_zero(self):
        assert isclose(
            sc.spherical_hankel2_d(0, 3), -0.3456774997623560 + 0.0629591636023160j
        )

    def test_non_zero(self):
        assert isclose(
            sc.spherical_hankel2_d(2, 3 + 4j), 2.193962713943190 - 5.312458889454144j
        )


class TestIntkambe:
    def test_m3_1(self):
        assert isclose(sc.intkambe(-3, 0.8, 1.2), 0.1536931539507005)

    def test_m3_2(self):
        assert isclose(sc.intkambe(-3, 0.3, 0.3), 256.2213077991314, rel_tol=1e-5)

    def test_m3_3(self):
        assert isclose(sc.intkambe(-3, 5, 4), 2.208658257800406e-91)

    def test_m3_complex(self):
        assert isclose(
            sc.intkambe(-3, 1 + 1j, 2 - 1.1j),
            -0.00009809435152495933 - 0.00015139494926761342j,
        )

    def test_m2_1(self):
        assert isclose(sc.intkambe(-2, 0.8, 1.2), 0.2340296298080376)

    def test_m2_2(self):
        assert isclose(sc.intkambe(-2, 0.3, 0.3), 87.3244612483742, rel_tol=1e-5)

    def test_m2_3(self):
        assert isclose(sc.intkambe(-2, 5, 4), 8.8564460016641e-91)

    def test_m2_complex(self):
        assert isclose(
            sc.intkambe(-2, 1 + 1j, 2 - 1.1j),
            -0.00039153718178222677 - 0.00019654173190154122j,
        )

    def test_m1_1(self):
        assert isclose(sc.intkambe(-1, 0.8, 1.2), 0.373562480945606)

    def test_m1_2(self):
        assert isclose(sc.intkambe(-1, 0.3, 0.3), 31.60318167885612, rel_tol=1e-5)

    def test_m1_3(self):
        assert isclose(sc.intkambe(-1, 5, 4), 3.55134657134584e-90)

    def test_m1_complex(self):
        assert isclose(
            sc.intkambe(-1, 1 + 1j, 2 - 1.1j),
            -0.001064442613991556 + 0.00007284003446857335j,
        )

    def test_0_1(self):
        assert isclose(sc.intkambe(0, 0.8, 1.2), 0.6329973232058413)

    def test_0_2(self):
        assert isclose(sc.intkambe(0, 0.3, 0.3), 14.50548439475395)

    def test_0_3(self):
        assert isclose(sc.intkambe(0, 5, 4), 1.424063216552536e-89)

    def test_0_complex(self):
        assert isclose(
            sc.intkambe(0, 1 + 1j, 2 - 1.1j),
            -0.002143348723888108 + 0.0014826182600382856j,
        )

    def test_fodd_1(self):
        assert isclose(sc.intkambe(9, 0.8, 1.2), 3722.587631513942)

    def test_fodd_2(self):
        assert isclose(sc.intkambe(9, 0.4, 0.4), 3.698976253172273e6, rel_tol=1e-5)

    def test_fodd_r(self):
        assert isclose(sc.intkambe(9, 5, 4), 3.818351328744268e-84)

    def test_fodd_complex(self):
        assert isclose(
            sc.intkambe(9, 1.4 + 0.01j, 3 - 0.01j),
            0.8489489550578334 - 0.09535673649786904j,
            rel_tol=1e-7,
        )

    def test_feven_1(self):
        assert isclose(sc.intkambe(10, 0.8, 1.2), 14289.0024493579)

    def test_feven_2(self):
        assert isclose(sc.intkambe(10, 0.4, 0.4), 2.849030924828978e7)

    def test_feven_3(self):
        assert isclose(sc.intkambe(10, 5, 4), 1.53122554488494e-83)

    def test_feven_complex(self):
        assert isclose(
            sc.intkambe(10, 1.4 + 0.01j, 3 - 0.01j),
            2.7511557962581623 - 0.3203862171927444j,
            rel_tol=1e-5,
        )

    @pytest.mark.skip
    def test_error(self):
        with pytest.raises(ValueError):
            sc.intkambe(0.3, 1, 1)

    def test_inf(self):
        assert np.isinf(sc.intkambe(1, 1, 0))

    def test_m3_zero(self):
        assert isclose(
            sc.intkambe(-3, 0, 1 + 1j), -0.031087578289355378 - 0.24740395925452308j
        )

    def test_m2_zero(self):
        assert isclose(
            sc.intkambe(-2, 0, 1 + 1j), 0.4554030049462477 - 0.5383650534833417j
        )

    def test_inf_2(self):
        assert np.isinf(sc.intkambe(-1, 0, 1))

    def test_inf_3(self):
        assert np.isinf(sc.intkambe(-1, 1, 0))

    def test_bodd_1(self):
        assert isclose(sc.intkambe(-9, 0.8, 1.2), 0.021334939311674394170463156)

    def test_bodd_2(self):
        assert isclose(sc.intkambe(-9, 0.4, 0.4), 2552.5567756521395093849138)

    def test_bodd_r(self):
        assert isclose(sc.intkambe(-9, 3, 2), 1.471110396630751e-12)

    def test_bodd_complex(self):
        assert isclose(
            sc.intkambe(-9, 1.4 + 0.01j, 3 - 0.01j),
            8.77953e-10 - 3.97705e-11j,
            rel_tol=1e-5,
        )

    def test_beven_1(self):
        assert isclose(sc.intkambe(-10, 0.8, 1.2), 0.0161459, rel_tol=1e-5)

    def test_beven_2(self):
        assert isclose(sc.intkambe(-10, 0.4, 0.4), 5915.31, rel_tol=1e-5)

    def test_beven_3(self):
        assert isclose(sc.intkambe(-10, 3, 2), 7.203365170219268e-13)

    def test_beven_complex(self):
        assert isclose(
            sc.intkambe(-10, 1.4 + 0.01j, 3 - 0.01j),
            2.82715e-10 - 1.18144e-11j,
            rel_tol=1e-5,
        )

    def test_m4_zzero(self):
        assert isclose(sc.intkambe(-4, 0, 3), 0.01276548573157296882)

    def test_m6_negz(self):
        assert isclose(
            sc.intkambe(-6, -2 - 1j, 3), 1.20189537847431e-10 + 3.80731298663362e-12j
        )

    def test_6_negz(self):
        assert isclose(sc.intkambe(6, -1, 3), 4.959558569294379)

    def test_m5_negz(self):
        assert isclose(
            sc.intkambe(-5, -2 + 1j, 3), 3.65682124179e-10 - 6.38779831117841e-12j
        )


class TestWignerD:
    @pytest.mark.skip
    def test_error(self):
        with pytest.raises(ValueError):
            sc.wignersmalld(-1, 0, 0, 0)

    def test_warning(self):
        assert sc.wignersmalld(1, 0, 2, 0) == 0

    def test_1(self):
        assert sc.wignersmalld(1, 0, 0, 0) == 1

    def test_2(self):
        assert sc.wignersmalld(1, 1, 0, 2 * np.pi) == 0

    def test_3(self):
        assert sc.wignersmalld(2, 1, -1, np.pi) == -1

    def test_4(self):
        assert sc.wignersmalld(2, 1, 0, 2 * np.pi) == 0

    def test_5(self):
        assert isclose(sc.wignersmalld(4, 3, 2, 1), -0.07526360176530718)

    def test_6(self):
        assert isclose(sc.wignersmalld(16, 8, -4, 2), -0.06370185806824848)

    def test_7(self):
        assert isclose(sc.wignersmalld(15, -7, -4, 4), -0.3126274668164052)

    def test_8(self):
        assert isclose(sc.wignersmalld(2, -1, 1, 6), 0.05815816395893696)

    def test_9(self):
        assert isclose(
            sc.wignerd(4, 2, 3, 1.0, 2.0, 3.0),
            -0.0011756123083512 - 0.2656305961739311j,
        )

    def test_9_complex(self):
        assert isclose(
            sc.wignerd(4, 2, 3, 1.0, 2j, 3.0), -250.9892374159303 + 1.1108134417492j
        )

    def test_10(self):
        assert isclose(
            sc.wignersmalld(2, 0, -2, 7 + 1j), 0.14867417641112 + 1.1000641894940703j
        )

    def test_11(self):
        assert isclose(sc.wignersmalld(2, 0, -2, np.pi + 1e-18), 0)


class TestPiFun:
    def test_1_m1_0(self):
        assert sc.pi_fun(1, -1, 1) == -0.5

    def test_2_1_pi(self):
        assert sc.pi_fun(3, 1, -1) == -6

    def test_3_3_0(self):
        assert sc.pi_fun(3, 3, 1) == 0

    def test_4_2_complex(self):
        assert isclose(
            sc.pi_fun(4, 2, np.cos(1 + 1j)), 51.9022970181194 - 253.1942074716320j
        )

    def test_nan_l(self):
        assert np.isnan(sc.pi_fun(0.5, 0, 1j))

    def test_nan_m(self):
        assert np.isnan(sc.pi_fun(1, 0.5, 1j))


class TestTauFun:
    def test_1_m1_0(self):
        assert sc.tau_fun(1, -1, 1) == 0.5

    def test_2_1_pi(self):
        assert isclose(sc.tau_fun(3, 3, np.cos(0.2)), -1.740723332980088)

    def test_4_2_complex(self):
        assert isclose(
            sc.tau_fun(4, 2, np.cos(1 + 1j)), -568.1643037179911 - 456.9245589672897j
        )

    def test_nan_l(self):
        assert np.isnan(sc.tau_fun(0.5, 0, 1j))

    def test_nan_m(self):
        assert np.isnan(sc.tau_fun(1, 0.5, 1j))


class TestVshZ:
    def test_0(self):
        assert np.array_equal(sc.vsh_Z(0, 0, 0, 0), [1j * np.sqrt(0.25 / np.pi), 0, 0])

    def test_1(self):
        assert np.array_equal(
            sc.vsh_Z(4, 3, 2j, 1), [1j * sc.sph_harm(3, 4, 1, 2j), 0, 0]
        )


class TestVshX:
    def test_0(self):
        assert np.array_equal(sc.vsh_X(0, 0, 0, 0), [0, 0, 0])

    def test_1(self):
        expect = (
            -1j
            * np.sqrt(15 / (2 * 3 * 8 * np.pi))
            * np.exp(1j)
            * np.array([0, 1j * np.cos(2j), -np.cos(4j)])
        )
        assert np.sum(np.abs(sc.vsh_X(2, 1, 2j, 1) - expect)) < EPS


class TestVshY:
    def test_0(self):
        assert np.array_equal(sc.vsh_Y(0, 0, 0, 0), [0, 0, 0])

    def test_1(self):
        expect = (
            -1j
            * np.sqrt(15 / (2 * 3 * 8 * np.pi))
            * np.exp(1j)
            * np.array([0, np.cos(4j), 1j * np.cos(2j)])
        )
        assert np.sum(np.abs(sc.vsh_Y(2, 1, 2j, 1) - expect)) < EPS


class TestVswM:
    def test_0(self):
        expect = (
            -1j
            * (ssc.spherical_jn(2, 4 + 1j) + 1j * ssc.spherical_yn(2, 4 + 1j))
            * np.sqrt(15 / (2 * 3 * 8 * np.pi))
            * np.exp(1j)
            * np.array([0, 1j * np.cos(2), -np.cos(4)])
        )
        assert np.sum(np.abs(sc.vsw_M(2, 1, 4 + 1j, 2, 1) - expect)) < EPS


class TestVswrM:
    def test_0(self):
        assert np.array_equal(sc.vsw_rM(1, 0, 0, 0, 0), [0, 0, 0])

    def test_1(self):
        expect = (
            -1j
            * ssc.spherical_jn(2, 4 + 1j)
            * np.sqrt(15 / (2 * 3 * 8 * np.pi))
            * np.exp(1j)
            * np.array([0, 1j * np.cos(2), -np.cos(4)])
        )
        assert np.sum(np.abs(sc.vsw_rM(2, 1, 4 + 1j, 2, 1) - expect)) < EPS


class TestVswN:
    def test_0(self):
        expect = (
            -1j
            * np.sqrt(5 / (16 * np.pi))
            * np.exp(1j)
            * np.array(
                [
                    6
                    * sc.spherical_hankel1(2, 4 + 1j)
                    / (4 + 1j)
                    * np.sin(2)
                    * np.cos(2),
                    np.cos(4)
                    * (
                        sc.spherical_hankel1_d(2, 4 + 1j)
                        + sc.spherical_hankel1(2, 4 + 1j) / (4 + 1j)
                    ),
                    1j
                    * np.cos(2)
                    * (
                        sc.spherical_hankel1_d(2, 4 + 1j)
                        + sc.spherical_hankel1(2, 4 + 1j) / (4 + 1j)
                    ),
                ]
            )
        )
        assert np.sum(np.abs(sc.vsw_N(2, 1, 4 + 1j, 2, 1) - expect)) < EPS


class TestVswrN:
    def test_0(self):
        assert (
            np.sum(
                np.abs(
                    sc.vsw_rN(1, 0, 0, 1, 0)
                    - [
                        1j / np.sqrt(6 * np.pi) * np.cos(1),
                        -1j / np.sqrt(6 * np.pi) * np.sin(1),
                        0,
                    ]
                )
            )
            < EPSSQ
        )

    def test_1(self):
        expect = (
            -1j
            * np.sqrt(5 / (16 * np.pi))
            * np.exp(1j)
            * np.array(
                [
                    6 * ssc.spherical_jn(2, 4 + 1j) / (4 + 1j) * np.sin(2) * np.cos(2),
                    np.cos(4)
                    * (
                        ssc.spherical_jn(2, 4 + 1j, 1)
                        + ssc.spherical_jn(2, 4 + 1j) / (4 + 1j)
                    ),
                    1j
                    * np.cos(2)
                    * (
                        ssc.spherical_jn(2, 4 + 1j, 1)
                        + ssc.spherical_jn(2, 4 + 1j) / (4 + 1j)
                    ),
                ]
            )
        )
        assert np.sum(np.abs(sc.vsw_rN(2, 1, 4 + 1j, 2, 1) - expect)) < EPS


class TestVswA:
    def test_p(self):
        assert np.array_equal(
            sc.vsw_A(5, 4, 3, 2, 1, 1),
            (sc.vsw_N(5, 4, 3, 2, 1) + sc.vsw_M(5, 4, 3, 2, 1)) * np.sqrt(0.5),
        )

    def test_m(self):
        assert np.array_equal(
            sc.vsw_A(5, 4, 3, 2, 1, 0),
            (sc.vsw_N(5, 4, 3, 2, 1) - sc.vsw_M(5, 4, 3, 2, 1)) * np.sqrt(0.5),
        )


class TestVswrA:
    def test_p(self):
        assert np.array_equal(
            sc.vsw_rA(5, 4, 3, 2, 1, 1),
            (sc.vsw_rN(5, 4, 3, 2, 1) + sc.vsw_rM(5, 4, 3, 2, 1)) * np.sqrt(0.5),
        )

    def test_m(self):
        assert np.array_equal(
            sc.vsw_rA(5, 4, 3j, 2, 1, 0),
            (sc.vsw_rN(5, 4, 3j, 2, 1) - sc.vsw_rM(5, 4, 3j, 2, 1)) * np.sqrt(0.5),
        )


class TestTlVswA:
    def test_0(self):
        assert isclose(
            sc.tl_vsw_A(1, 0, 1, 0, 1, 0, 0), 0.903506036819270 - 4.145319872028107j
        )

    def test_1(self):
        assert isclose(
            sc.tl_vsw_A(14, 13, 11, -3, 12, 2, 1),
            -6.265680341371548e02 - 2.084168034177037e03j,
        )


class TestTlVswrA:
    def test_0(self):
        assert isclose(sc.tl_vsw_rA(1, 0, 1, 0, 1, 0, 0), 0.903506036819270)

    def test_1(self):
        assert isclose(
            sc.tl_vsw_rA(14, 13, 11, -3, 12 + 0j, 2, 1),
            6.853762595651612e-05 - 2.060462015430312e-05j,
        )


class TestTlVswB:
    def test_0(self):
        assert isclose(sc.tl_vsw_B(1, 0, 1, 0, 1, 0, 0), 0)

    def test_1(self):
        assert isclose(
            sc.tl_vsw_B(14, 13, 11, -3, 12, 2, 1),
            -2.087226903652359e02 + 6.274791216231195e01j,
        )


class TestTlVswrB:
    def test_0(self):
        assert isclose(sc.tl_vsw_rB(1, 0, 1, 0, 1, 0, 0), 0)

    def test_1(self):
        assert isclose(
            sc.tl_vsw_rB(14, 13, 11, -3, 12, 2, 1),
            -2.366181401015062e-04 - 7.870684079277587e-04j,
        )


class TestVcwM:
    def test_0(self):
        assert np.array_equal(sc.vcw_M(0, 0, 1, 0, 0), [0, -sc.hankel1_d(0, 1), 0])

    def test_1(self):
        assert (
            np.sum(
                np.abs(
                    sc.vcw_M(1, -2, 3, 4, 5)
                    - np.exp(-3j)
                    * np.array([-2j / 3 * sc.hankel1(-2, 3), -sc.hankel1_d(-2, 3), 0])
                )
            )
            < EPSSQ
        )


class TestVcwrM:
    def test_0(self):
        assert np.array_equal(sc.vcw_rM(0, 0, 1, 0, 0), [0, -sc.jv_d(0, 1), 0])

    def test_1(self):
        assert (
            np.sum(
                np.abs(
                    sc.vcw_rM(1, -2, 3, 4, 5)
                    - np.exp(-3j)
                    * np.array([-2j / 3 * sc.jv(-2, 3), -sc.jv_d(-2, 3), 0])
                )
            )
            < EPSSQ
        )

    def test_origin_m0(self):
        assert np.array_equal(sc.vcw_rM(0, 0, 0, 0, 0), [0, 0, 0])

    def test_origin_m1(self):
        assert np.array_equal(sc.vcw_rM(0, 1, 0, 0, 0), [0.5j, -0.5, 0])

    def test_origin_m4(self):
        assert np.array_equal(sc.vcw_rM(0, 4, 0, 0, 0), [0, 0, 0])


class TestVcwN:
    def test_0(self):
        assert np.array_equal(sc.vcw_N(0, 0, 1, 0, 0, 1), [0, 0, sc.hankel1(0, 1)])

    def test_1(self):
        assert (
            np.sum(
                np.abs(
                    sc.vcw_N(4, -2, 3, 1, 2, 5)
                    - np.exp(6j)
                    * np.array(
                        [
                            4j / 5 * sc.hankel1_d(-2, 3),
                            2 * 4 / (3 * 5) * sc.hankel1(-2, 3),
                            3 / 5 * sc.hankel1(-2, 3),
                        ]
                    )
                )
            )
            < EPSSQ
        )


class TestVcwrN:
    def test_0(self):
        assert np.array_equal(sc.vcw_rN(0, 0, 1, 0, 0, 1), [0, 0, sc.jv(0, 1)])

    def test_1(self):
        assert (
            np.sum(
                np.abs(
                    sc.vcw_rN(4, -2, 3, 1, 2, 5)
                    - np.exp(6j)
                    * np.array(
                        [
                            4j / 5 * sc.jv_d(-2, 3),
                            2 * 4 / (3 * 5) * sc.jv(-2, 3),
                            3 / 5 * sc.jv(-2, 3),
                        ]
                    )
                )
            )
            < EPSSQ
        )


class TestVcwA:
    def test_p(self):
        assert np.array_equal(
            sc.vcw_A(4, -2, 3, 1, 2, 5, 1),
            (sc.vcw_N(4, -2, 3, 1, 2, 5) + sc.vcw_M(4, -2, 3, 1, 2)) * np.sqrt(0.5),
        )

    def test_m(self):
        assert np.array_equal(
            sc.vcw_A(4, -2, 3, 1, 2, 5, 0),
            (sc.vcw_N(4, -2, 3, 1, 2, 5) - sc.vcw_M(4, -2, 3, 1, 2)) * np.sqrt(0.5),
        )


class TestVcwrA:
    def test_p(self):
        assert np.array_equal(
            sc.vcw_rA(4, -2, 3, 1, 2, 5, 1),
            (sc.vcw_rN(4, -2, 3, 1, 2, 5) + sc.vcw_rM(4, -2, 3, 1, 2)) * np.sqrt(0.5),
        )

    def test_m(self):
        assert np.array_equal(
            sc.vcw_rA(4, -2, 3, 1, 2, 5, 0),
            (sc.vcw_rN(4, -2, 3, 1, 2, 5) - sc.vcw_rM(4, -2, 3, 1, 2)) * np.sqrt(0.5),
        )


class TestTlVcw:
    def test_0(self):
        assert sc.tl_vcw(0, 0, 1, 0, 0, 0, 0) == 0

    def test_1(self):
        assert sc.tl_vcw(1, 2, 1, 3, 4, 5, 6) == np.exp(11j) * sc.hankel1(1, 4)


class TestTlVcwr:
    def test_0(self):
        assert sc.tl_vcw_r(0, 0, 1, 0, 0, 0, 0) == 0

    def test_1(self):
        assert sc.tl_vcw_r(1, 2, 1, 3, 4, 5, 6) == np.exp(11j) * sc.jv(1, 4)


class TestVpwM:
    def test_k0(self):
        res = sc.vpw_M(0, 0, 0j, 0, 0, 0)
        assert np.all([np.isnan(i) for i in res])

    def test_kpar0(self):
        assert np.array_equal(sc.vpw_M(0, 0, 3, 3, 2, 1), [0, -1j * np.exp(3j), 0])

    def test_k_general(self):
        k = np.array([3, 4, -3])
        r = np.array([4, 0.2, 3])
        res = sc.vpw_M(*k, *r)
        assert np.all(
            np.abs(
                res
                - [
                    1j * k[1] * np.exp(1j * k @ r) / 5,
                    -1j * k[0] * np.exp(1j * k @ r) / 5,
                    0,
                ]
            )
            < EPSSQ
        )


class TestVpwN:
    def test_k0(self):
        res = sc.vpw_N(0, 0, 0j, 0, 0, 0)
        assert np.all([np.isnan(i) for i in res])

    def test_kpar0(self):
        assert np.array_equal(sc.vpw_N(0, 0, 3, 3, 2, 1), [-np.exp(3j), 0, 0])

    def test_kpar0_imagkz(self):
        assert np.array_equal(sc.vpw_N(0, 0, -3j, 3, 2, 1), [np.exp(3), 0, 0])

    def test_k_general(self):
        k = np.array([3, 4, -3])
        r = np.array([4, 0.2, 3])
        res = sc.vpw_N(*k, *r)
        assert np.all(
            np.abs(
                res
                - [
                    -0.2441697182761140 - 0.1888789726889204j,
                    -0.3255596243681520 - 0.2518386302518939j,
                    -0.6782492174336500 - 0.5246638130247790j,
                ]
            )
            < EPSSQ
        )


class TestVpwA:
    def test(self):
        k = np.array([-3, 1, -6])
        r = np.array([2, 0.2, 0.5])
        assert np.all(
            np.abs(
                np.sqrt(2) * sc.vpw_A(*k, *r, 0) - sc.vpw_N(*k, *r) + sc.vpw_M(*k, *r)
            )
            < EPSSQ
        )

    def test_complex(self):
        k = np.array([-3, 1j, -6])
        r = np.array([2, 0.2, 0.5])
        assert np.all(
            np.abs(
                np.sqrt(2) * sc.vpw_A(*k, *r, 0) - sc.vpw_N(*k, *r) + sc.vpw_M(*k, *r)
            )
            < EPSSQ
        )


class TestCar2Cyl:
    def test_0(self):
        assert np.array_equal(sc.car2cyl([0, 0, 0]), [0, 0, 0])

    def test_1(self):
        assert np.array_equal(sc.car2cyl([3, -4, 1]), [5, -0.9272952180016122, 1])


class TestCar2Sph:
    def test_0(self):
        assert np.array_equal(sc.car2sph([0, 0, 0]), [0, 0, 0])

    def test_1(self):
        assert np.all(
            np.abs(
                sc.car2sph([3, -4, 1])
                - [np.sqrt(26), np.arctan2(5, 1), -0.9272952180016122]
            )
            < EPSSQ
        )


class TestCyl2Car:
    def test_0(self):
        assert np.array_equal(sc.cyl2car([0, 1, 0]), [0, 0, 0])

    def test_1(self):
        assert (
            np.sum(np.abs(sc.cyl2car([5, -0.9272952180016122, 1]) - [3, -4, 1])) < EPSSQ
        )


class TestCyl2Sph:
    def test_0(self):
        assert np.array_equal(sc.cyl2sph([0, 1, 0]), [0, 0, 1])

    def test_1(self):
        assert (
            np.sum(np.abs(sc.cyl2sph([3, -4, 1]) - [np.sqrt(10), np.arctan2(3, 1), -4]))
            < EPSSQ
        )


class TestSph2Car:
    def test_0(self):
        assert np.array_equal(sc.sph2car([0, 1, 2]), [0, 0, 0])

    def test_1(self):
        assert (
            np.sum(
                np.abs(
                    sc.sph2car([np.sqrt(26), np.arctan2(5, 1), -0.9272952180016122])
                    - [3, -4, 1]
                )
            )
            < EPSSQ
        )


class TestSph2Cyl:
    def test_0(self):
        assert np.array_equal(sc.sph2cyl([0, 0, 1]), [0, 1, 0])

    def test_1(self):
        assert (
            np.sum(np.abs(sc.sph2cyl([np.sqrt(10), np.arctan2(3, 1), -4]) - [3, -4, 1]))
            < EPSSQ
        )


class TestCar2Pol:
    def test_0(self):
        assert np.array_equal(sc.car2pol([0, 0]), [0, 0])

    def test_1(self):
        assert np.array_equal(sc.car2pol([3, -4]), [5, -0.9272952180016122])


class TestPol2Car:
    def test_0(self):
        assert np.array_equal(sc.pol2car([0, 1]), [0, 0])

    def test_1(self):
        assert np.sum(np.abs(sc.pol2car([5, -0.9272952180016122]) - [3, -4])) < EPSSQ


class TestVCarCyl:
    def test_0(self):
        assert np.array_equal(sc.vcar2cyl([0, 0, 0j], [0, 0, 0]), [0, 0, 0])

    def test_1(self):
        assert np.array_equal(sc.vcar2cyl([1, 0, 0], [0, 0, 0]), [1, 0, 0])

    def test_2(self):
        assert np.array_equal(sc.vcyl2car([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    def test_3(self):
        assert np.array_equal(sc.vcyl2car([1, 0, 0], [0, 0, 0]), [1, 0, 0])

    def test_4(self):
        pcar = [3, -4, 1]
        vcar = [1, 2, 3]
        pcyl = sc.car2cyl(pcar)
        vcyl = sc.vcar2cyl(vcar, pcar)
        assert np.sum(np.abs(sc.vcyl2car(vcyl, pcyl) - vcar)) < EPSSQ


class TestVCarSph:
    def test_0(self):
        assert np.array_equal(sc.vcar2sph([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    def test_1(self):
        assert np.array_equal(sc.vcar2sph([1, 0, 0], [0, 0, 0]), [0, 1, 0])

    def test_2(self):
        assert np.array_equal(sc.vsph2car([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    def test_3(self):
        assert np.array_equal(sc.vsph2car([0, 1, 0], [0, 0, 0]), [1, 0, 0])

    def test_4(self):
        pcar = [3, -4, 1]
        vcar = [1, 2, 3]
        psph = sc.car2sph(pcar)
        vsph = sc.vcar2sph(vcar, pcar)
        assert np.sum(np.abs(sc.vsph2car(vsph, psph) - vcar)) < EPSSQ


class TestVCylSph:
    def test_0(self):
        assert np.array_equal(sc.vcyl2sph([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    def test_1(self):
        assert np.array_equal(sc.vcyl2sph([1, 0, 0], [0, 0, 0]), [0, 1, 0])

    def test_2(self):
        assert np.array_equal(sc.vsph2cyl([0, 0, 0], [0, 0, 0]), [0, 0, 0])

    def test_3(self):
        assert np.array_equal(sc.vsph2cyl([0, 1, 0], [0, 0, 0]), [1, 0, 0])

    def test_4(self):
        pcar = [3, -4, 1]
        vcar = [1, 2, 3]
        psph = sc.cyl2sph(pcar)
        vsph = sc.vcyl2sph(vcar, pcar)
        assert np.sum(np.abs(sc.vsph2cyl(vsph, psph) - vcar)) < EPSSQ


class TestVCarPol:
    def test_0(self):
        assert np.array_equal(sc.vcar2pol([0, 0], [0, 0]), [0, 0])

    def test_1(self):
        assert np.array_equal(sc.vcar2pol([1, 0], [0, 0]), [1, 0])

    def test_2(self):
        assert np.array_equal(sc.vpol2car([0, 0], [0, 0]), [0, 0])

    def test_3(self):
        assert np.array_equal(sc.vpol2car([1, 0], [0, 0]), [1, 0])

    def test_4(self):
        pcar = [3, -4]
        vcar = [1, 2]
        pcyl = sc.car2pol(pcar)
        vcyl = sc.vcar2pol(vcar, pcar)
        assert np.sum(np.abs(sc.vpol2car(vcyl, pcyl) - vcar)) < EPSSQ


class TestCythonSpecial:
    def test_hankel1_d(self):
        assert sc.hankel1_d(0.4, 0.3j) == cs.hankel1_d(0.4, 0.3j)

    def test_hankel2_d(self):
        assert sc.hankel2_d(0.4, 0.3j) == cs.hankel2_d(0.4, 0.3j)

    def test_jv_d(self):
        assert sc.jv_d(0.4, 0.3j) == cs.jv_d(0.4, 0.3j)

    def test_spherical_hankel1(self):
        assert sc.spherical_hankel1(4, 0.3j) == cs.spherical_hankel1(4, 0.3j)

    def test_spherical_hankel2(self):
        assert sc.spherical_hankel2(4, 0.3j) == cs.spherical_hankel2(4, 0.3j)

    def test_spherical_hankel1_d(self):
        assert sc.spherical_hankel1_d(4, 0.3j) == cs.spherical_hankel1_d(4, 0.3j)

    def test_spherical_hankel2_d(self):
        assert sc.spherical_hankel2_d(4, 0.3j) == cs.spherical_hankel2_d(4, 0.3j)

    def test_yv_d(self):
        assert sc.yv_d(0.4, 0.3j) == cs.yv_d(0.4, 0.3j)

    def test_incgamma(self):
        assert sc.incgamma(1.5, 0.3j) == cs.incgamma(1.5, 0.3j)

    def test_intkambe(self):
        assert sc.intkambe(4, 0.3j, 0.2 + 1j) == cs.intkambe(4, 0.3j, 0.2 + 1j)

    def test_lpmv(self):
        assert sc.lpmv(3, 4, 0.2 + 1j) == cs.lpmv(3, 4, 0.2 + 1j)

    def test_pi_fun(self):
        assert sc.pi_fun(4, 3, 0.2 + 1j) == cs.pi_fun(4, 3, 0.2 + 1j)

    def test_sph_harm(self):
        assert sc.sph_harm(3, 4, 1, 0.2 + 1j) == cs.sph_harm(3, 4, 1, 0.2 + 1j)

    def test_tau_fun(self):
        assert sc.tau_fun(4, 3, 0.2 + 1j) == cs.tau_fun(4, 3, 0.2 + 1j)

    def test_tl_vcw(self):
        assert sc.tl_vcw(6, 5, 4, 3, 2, 1, 0) == cs.tl_vcw(6, 5, 4, 3, 2, 1, 0)

    def test_tl_vcw_r(self):
        assert sc.tl_vcw_r(6, 5, 4, 3, 2.0, 1, 0) == cs.tl_vcw_r(6, 5, 4, 3, 2.0, 1, 0)

    def test_tl_vsw_A(self):
        assert sc.tl_vsw_A(6, 5, 4, 3, 2, 1.0, 0) == cs.tl_vsw_A(6, 5, 4, 3, 2, 1.0, 0)

    def test_tl_vsw_rA(self):
        assert sc.tl_vsw_rA(6, 5, 4, 3, 2, 1.0, 0) == cs.tl_vsw_rA(
            6, 5, 4, 3, 2.0, 1.0, 0
        )

    def test_tl_vsw_B(self):
        assert sc.tl_vsw_B(6, 5, 4, 3, 2, 1.0, 0) == cs.tl_vsw_B(6, 5, 4, 3, 2, 1.0, 0)

    def test_tl_vsw_rB(self):
        assert sc.tl_vsw_rB(6, 5, 4, 3, 2, 1.0, 0) == cs.tl_vsw_rB(
            6, 5, 4, 3, 2.0, 1.0, 0
        )

    def test_wigner3j(self):
        assert sc.wigner3j(5, 4, 3, -3, 2, 1) == cs.wigner3j(5, 4, 3, -3, 2, 1)

    def test_wigner3j(self):
        assert sc.wigner3j(5, 4, 3, -3, 2, 1) == cs.wigner3j(5, 4, 3, -3, 2, 1)

    def test_wignersmalld(self):
        assert sc.wignersmalld(5, 4, 3, 2) == cs.wignersmalld(5, 4, 3, 2.0)

    def test_wignerd(self):
        assert sc.wignerd(5, 4, 3, 2, 1, 0) == cs.wignerd(5, 4, 3, 2, 1.0, 0)



import numpy as np
import pytest

import treams.lattice as la


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def osscilation_avg(res):
    avg_len = len(res) // 2
    return np.mean(np.cumsum(res)[-avg_len:])


EPS = 2e-7
EPSSQ = 4e-14


class TestLSumSW1d:
    @pytest.mark.parametrize("l", [0, 1, 2, 7, 8])
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [-0.2, 0, 0.7])
    @pytest.mark.parametrize("a", [1, 1.1])
    @pytest.mark.parametrize("r", [-0.2, 0, 0.5])
    def test_regular(self, l, k, kpar, a, r):  # noqa: E741
        assert isclose(
            la.lsumsw1d(l, k, kpar, a, r, 0),
            np.sum(la.dsumsw1d(l, k, kpar, a, r, np.arange(100_000))),
            rel_tol=0.02,
            abs_tol=EPSSQ,
        )

    @pytest.mark.parametrize("l", [0, 1, 2, 7, 8])
    @pytest.mark.parametrize("r", [-0.2, 0, 0.5])
    @pytest.mark.slow
    def test_singular(self, l, r):  # noqa: E741
        a = 2 * np.pi
        k = 1
        kpar = 0
        assert isclose(
            la.lsumsw1d(l, k, kpar, a, r, 0),
            np.sum(la.dsumsw1d(l, k, kpar, a, r, np.arange(1_100_000))),
            rel_tol=0.05,
            abs_tol=EPSSQ,
        )


class TestLSumCW1d:
    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j, 2j])
    @pytest.mark.parametrize("kpar", [-0.2, 0, 0.7])
    @pytest.mark.parametrize("a", [1, 1.1])
    @pytest.mark.parametrize("r", [-0.2, 0, 0.5])
    def test_regular(self, l, k, kpar, a, r):  # noqa: E741
        assert isclose(
            la.lsumcw1d(l, k, kpar, a, r, 0),
            np.sum(la.dsumcw1d(l, k, kpar, a, r, np.arange(100_000))),
            rel_tol=0.02,
            abs_tol=EPSSQ,
        )

    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("r", [-0.2, 0, 0.5])
    def test_singular(self, l, r):  # noqa: E741
        a = 2 * np.pi
        k = 1
        kpar = 0
        assert isclose(
            la.lsumcw1d(l, k, kpar, a, r, 0),
            np.sum(la.dsumcw1d(l, k, kpar, a, r, np.arange(1_500_000))),
            rel_tol=0.05,
            abs_tol=EPSSQ,
        )


class TestLSumCW2d:
    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [np.zeros(2), np.array([0.1, 0.2])])
    @pytest.mark.parametrize(
        "a", [np.array([[1, 0], [0, 1.2]]), np.array([[1.1, 0.2], [-0.1, 1]])]
    )
    @pytest.mark.parametrize("r", [np.zeros(2), np.array([0.2, 0])])
    @pytest.mark.slow
    def test_regular(self, l, k, kpar, a, r):  # noqa: E741
        assert isclose(
            la.lsumcw2d(l, k, kpar, a, r, 0),
            osscilation_avg(la.dsumcw2d(l, k, kpar, a, r, np.arange(800))),
            rel_tol=1e-1,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("r", [np.zeros(2), np.array([0.2, 0])])
    @pytest.mark.slow
    def test_singular(self, l, r):  # noqa: E741
        k = 1
        kpar = np.zeros(2)
        a = np.array([[2 * np.pi, 0], [0, 1]])
        # Here the direct result is too oscillatory
        if l == -2:  # noqa: E741
            if np.all(r == 0):
                assert isclose(
                    np.imag(la.lsumcw2d(l, k, kpar, a, r, 0)), 2.9105962449216736
                )
        elif l == 0:  # noqa: E741
            if np.all(r == 0):
                assert isclose(
                    np.imag(la.lsumcw2d(l, k, kpar, a, r, 0)), 0.9156690769853802
                )
            else:
                assert isclose(
                    np.imag(la.lsumcw2d(l, k, kpar, a, r, 0)), -0.15877332685629297
                )
        else:
            assert isclose(
                np.imag(la.lsumcw2d(l, k, kpar, a, r, 0)),
                np.imag(np.sum(la.dsumcw2d(l, k, kpar, a, r, np.arange(1000)))),
                rel_tol=1e-1,
                abs_tol=EPS,
            )


class TestLSumSW2d:
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [np.zeros(2), np.array([0.1, 0.2])])
    @pytest.mark.parametrize(
        "a", [np.array([[1, 0], [0, 1.2]]), np.array([[1.1, 0.2], [-0.1, 1]])]
    )
    @pytest.mark.parametrize("r", [np.zeros(2), np.array([0.2, 0])])
    @pytest.mark.slow
    def test_regular(self, lm, k, kpar, a, r):
        assert isclose(
            la.lsumsw2d(*lm, k, kpar, a, r, 0),
            osscilation_avg(la.dsumsw2d(*lm, k, kpar, a, r, np.arange(800))),
            rel_tol=1e-1,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize("lm", [(0, 0), (2, -2), (2, 0), (8, -8), (8, 0), (8, 4)])
    @pytest.mark.slow
    def test_singular_r0(self, lm):
        k = 1
        kpar = np.zeros(2)
        a = np.array([[2 * np.pi, 0], [0, 1]])
        r = np.zeros(2)
        dirres = np.cumsum(la.dsumsw2d(*lm, k, kpar, a, r, np.arange(800)))[[11, -1]]
        assert isclose(
            np.angle(la.lsumsw2d(*lm, k, kpar, a, r, 0) - dirres[0]),
            np.angle(dirres[1] - dirres[0]),
            rel_tol=1e-1,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize("lm", [(0, 0), (2, -2), (2, 0), (7, -7), (7, 3), (8, 0)])
    @pytest.mark.slow
    def test_singular_r1(self, lm):
        k = 1
        kpar = np.zeros(2)
        a = np.array([[2 * np.pi, 0], [0, 1]])
        r = np.array([0.2, 0])
        dirres = np.cumsum(la.dsumsw2d(*lm, k, kpar, a, r, np.arange(800)))[[11, -1]]
        assert isclose(
            np.angle(la.lsumsw2d(*lm, k, kpar, a, r, 0) - dirres[0]),
            np.angle(dirres[1] - dirres[0]),
            rel_tol=1e-1,
            abs_tol=EPS,
        )


class TestLSumSW3d:
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [np.zeros(3), np.array([0.1, 0.2, -0.3])])
    @pytest.mark.parametrize(
        "a",
        [
            np.array([[1, 0, 0], [0, 2.0, 0], [0, 0, 3]]),
            np.array([[1.3, 0.2, 0.1], [-0.1, 1.9, 0.2], [-0.1, 0.3, 2.2]]),
        ],
    )
    @pytest.mark.parametrize("r", [np.zeros(3), np.array([0.2, -0.3, 0.4])])
    @pytest.mark.slow
    def test_regular(self, lm, k, kpar, a, r):
        if lm[0] in (0, 2) and lm[1] == 0:
            rtol = 0.2
            if (
                lm[0] == 2
                and np.array_equal(kpar, [0.1, 0.2, -0.3])
                and np.array_equal(
                    a, np.array([[1.3, 0.2, 0.1], [-0.1, 1.9, 0.2], [-0.1, 0.3, 2.2]])
                )
            ):
                rtol = 0.3
        else:
            rtol = 0.1
        assert isclose(
            la.lsumsw3d(*lm, k, kpar, a, r, 0),
            osscilation_avg(la.dsumsw3d(*lm, k, kpar, a, r, np.arange(120))),
            rel_tol=rtol,
            abs_tol=1e-8,
        )

    # Only the imaginary part fits
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 1),
            (5, -5),
            (5, 0),
            (5, 2),
            (6, -6),
            (6, 0),
            (6, 3),
        ],
    )
    def test_singular_r0(self, lm):
        k = 1
        kpar = np.zeros(3)
        a = np.array([[2 * np.pi, 0, 0], [0, 2 * np.pi, 0], [0, 0, 3]])
        r = np.zeros(3)
        assert isclose(
            np.imag(la.lsumsw3d(*lm, k, kpar, a, r, 0)),
            np.imag(np.sum(la.dsumsw3d(*lm, k, kpar, a, r, np.arange(2)))),
            rel_tol=0.2,
            abs_tol=1e-8,
        )

    # Only the imaginary part fits
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (5, -5),
            (5, 0),
            (5, 2),
            (6, -6),
            (6, 0),
            (6, 3),
        ],
    )
    def test_singular_r1(self, lm):
        k = 1
        kpar = np.zeros(3)
        a = np.array([[2 * np.pi, 0, 0], [0, 2 * np.pi, 0], [0, 0, 3]])
        r = np.array([0.2, -0.3, 0.4])
        assert isclose(
            np.imag(la.lsumsw3d(*lm, k, kpar, a, r, 0)),
            np.imag(np.sum(la.dsumsw3d(*lm, k, kpar, a, r, np.arange(2)))),
            rel_tol=0.2,
            abs_tol=1e-8,
        )


class TestLSumSW2dShift:
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [np.zeros(2), np.array([0.1, 0.2])])
    @pytest.mark.parametrize(
        "a", [np.array([[1, 0], [0, 1.2]]), np.array([[1.1, 0.2], [-0.1, 1]])]
    )
    @pytest.mark.parametrize("r", [np.array([0, 0, 0.1]), np.array([0.2, 0.2, 1.1])])
    @pytest.mark.slow
    def test_regular(self, lm, k, kpar, a, r):
        assert isclose(
            la.lsumsw2d_shift(*lm, k, kpar, a, r, 0),
            osscilation_avg(la.dsumsw2d_shift(*lm, k, kpar, a, r, np.arange(800))),
            rel_tol=5e-2,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize(
        "lm",
        [(0, 0), (1, 0), (2, -2), (2, 0), (2, 1), (7, 0), (8, -8), (8, 0), (8, 4)],
    )
    @pytest.mark.slow
    def test_singular_r0(self, lm):
        r = np.array([0, 0, 0.1])
        k = 1
        kpar = np.zeros(2)
        a = np.array([[2 * np.pi, 0], [0, 1]])
        if (lm[0] + lm[1]) % 2 == 1:
            assert isclose(
                la.lsumsw2d_shift(*lm, k, kpar, a, r, 0),
                la.dsumsw2d_shift(*lm, k, kpar, a, r, np.arange(800)).sum(),
                rel_tol=1e-2,
                abs_tol=EPS,
            )
        else:
            dirres = np.cumsum(la.dsumsw2d_shift(*lm, k, kpar, a, r, np.arange(800)))[
                [1, -1]
            ]
            assert isclose(
                np.angle(la.lsumsw2d_shift(*lm, k, kpar, a, r, 0) - dirres[0]),
                np.angle(dirres[1] - dirres[0]),
                rel_tol=1e-1,
                abs_tol=EPS,
            )

    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.slow
    def test_singular_r1(self, lm):
        r = np.array([0.2, 0.2, 1.1])
        k = 1
        kpar = np.zeros(2)
        a = np.array([[2 * np.pi, 0], [0, 1]])
        if (lm[0] + lm[1]) % 2 == 1:
            assert isclose(
                la.lsumsw2d_shift(*lm, k, kpar, a, r, 0),
                la.dsumsw2d_shift(*lm, k, kpar, a, r, np.arange(1000)).sum(),
                rel_tol=1e-1,
                abs_tol=EPS,
            )
        else:
            dirres = np.cumsum(la.dsumsw2d_shift(*lm, k, kpar, a, r, np.arange(1000)))[
                [11, -1]
            ]
            assert isclose(
                np.angle(la.lsumsw2d_shift(*lm, k, kpar, a, r, 0) - dirres[0]),
                np.angle(dirres[1] - dirres[0]),
                rel_tol=1e-1,
                abs_tol=EPS,
            )


class TestLSumSW1dShift:
    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [-0.2, 0, 0.7])
    @pytest.mark.parametrize("a", [1, 1.1])
    @pytest.mark.parametrize("r", [np.array([0, 0.1, 0]), np.array([0.2, 1.1, 0.2])])
    def test_regular(self, lm, k, kpar, a, r):
        assert isclose(
            la.lsumsw1d_shift(*lm, k, kpar, a, r, 0),
            la.dsumsw1d_shift(*lm, k, kpar, a, r, np.arange(100_000)).sum(),
            rel_tol=1e-2,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize(
        "lm",
        [
            (0, 0),
            (1, -1),
            (1, 0),
            (2, -2),
            (2, 0),
            (2, 1),
            (7, -7),
            (7, 0),
            (7, 3),
            (8, -8),
            (8, 0),
            (8, 4),
        ],
    )
    @pytest.mark.parametrize("r", [np.array([0, 0.1, 0]), np.array([0.2, 0.2, 1.1])])
    @pytest.mark.slow
    def test_singular(self, lm, r):
        k = 1
        kpar = 0
        a = 2 * np.pi
        if lm == (2, 0) and np.array_equal(r, [0.2, 0.2, 1.1]):
            assert isclose(
                la.lsumsw1d_shift(*lm, k, kpar, a, r, 0),
                la.dsumsw1d_shift(*lm, k, kpar, a, r, np.arange(2_200_000)).sum(),
                rel_tol=1e-1,
                abs_tol=EPS,
            )
        else:
            assert isclose(
                la.lsumsw1d_shift(*lm, k, kpar, a, r, 0),
                la.dsumsw1d_shift(*lm, k, kpar, a, r, np.arange(800_000)).sum(),
                rel_tol=1e-1,
                abs_tol=EPS,
            )


class TestLSumCW1dShift:
    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("k", [2, 1.1 + 0.1j])
    @pytest.mark.parametrize("kpar", [-0.2, 0, 0.7])
    @pytest.mark.parametrize("a", [1, 1.1])
    @pytest.mark.parametrize("r", [np.array([0, 0.1]), np.array([0.2, 1.1])])
    def test_regular(self, l, k, kpar, a, r):  # noqa: E741
        assert isclose(
            la.lsumcw1d_shift(l, k, kpar, a, r, 0),
            la.dsumcw1d_shift(l, k, kpar, a, r, np.arange(100_000)).sum(),
            rel_tol=5e-2,
            abs_tol=EPS,
        )

    @pytest.mark.parametrize("l", [-2, -1, 0, 7, 8])
    @pytest.mark.parametrize("r", [np.array([0, 0.1]), np.array([0.2, 1.1])])
    @pytest.mark.slow
    def test_singular(self, l, r):  # noqa: E741
        k = 1
        kpar = 0
        a = 2 * np.pi
        assert isclose(
            la.lsumcw1d_shift(l, k, kpar, a, r, 0),
            la.dsumcw1d_shift(l, k, kpar, a, r, np.arange(1_800_000)).sum(),
            rel_tol=1e-1,
            abs_tol=EPS,
        )



import numpy as np
import pytest

from treams import Lattice, WaveVector


class TestLattice:
    def test_init(self):
        assert Lattice([1]) == Lattice(1, alignment="z")

    def test_init_lattice(self):
        assert Lattice(Lattice([1, 2])) == Lattice([[1, 0], [0, 2]])

    def test_init_invalid_shape(self):
        with pytest.raises(ValueError):
            Lattice([0, 1, 2, 3])

    def test_init_invalid_alignment(self):
        with pytest.raises(ValueError):
            Lattice(1, "a")

    def test_init_volume(self):
        with pytest.raises(ValueError):
            Lattice([[1, 2], [2, 4]])

    def test_square(self):
        assert Lattice.square(1) == Lattice([1, 1])

    def test_cubic(self):
        assert Lattice.cubic(1) == Lattice([1, 1, 1])

    def test_rectangular(self):
        assert Lattice.rectangular(1, 2) == Lattice([1, 2])

    def test_orthorhombic(self):
        assert Lattice.orthorhombic(1, 2, 3) == Lattice([1, 2, 3])

    def test_hexagonal_2d(self):
        assert Lattice.hexagonal(1) == Lattice([[1, 0], [0.5, np.sqrt(0.75)]])

    def test_hexagonal_3d(self):
        assert Lattice.hexagonal(1, 2) == Lattice(
            [[1, 0, 0], [0.5, 0.75 ** 0.5, 0], [0, 0, 2]]
        )

    def test_reciprocal(self):
        assert (Lattice([np.pi, 2 * np.pi]).reciprocal == [[2, 0], [0, 1]]).all()

    def test_str(self):
        assert str(Lattice(1)) == "1.0"

    def test_repr(self):
        assert (
            repr(Lattice([1, 2]))
            == """Lattice([[1. 0.]
         [0. 2.]], alignment='xy')"""
        )

    def test_sublattice(self):
        assert Lattice(Lattice([1, 2, 3]), "xy") == Lattice([1, 2])

    def test_sublattice_1d(self):
        assert Lattice(Lattice(1), "z") == Lattice(1)

    def test_sublattice_1d_error(self):
        with pytest.raises(ValueError):
            Lattice(Lattice(1), "x")

    def test_sublattice_2d_error(self):
        with pytest.raises(ValueError):
            Lattice(Lattice([1, 2], "zx"), "xy")

    def test_sublattice_find_error(self):
        with pytest.raises(ValueError):
            Lattice(Lattice([[1, 1], [0, 1]]), "x")

    def test_permute(self):
        assert Lattice([1, 2]).permute(4) == Lattice([1, 2], "yz")

    def test_permute_3d(self):
        assert Lattice([1, 2, 3]).permute(2.0) == Lattice(
            [[0, 0, 1], [2, 0, 0], [0, 3, 0]]
        )

    def test_permute_error(self):
        with pytest.raises(ValueError):
            Lattice(1).permute(0.2)

    def test_bool(self):
        assert bool(Lattice(1))

    def test_or(self):
        assert Lattice(1) | Lattice([2, 3]) == Lattice([2, 3, 1])

    def test_or_none(self):
        assert Lattice(1) | None == Lattice(1)

    def test_or_1d(self):
        assert Lattice(1) | Lattice(2, "x") == Lattice([1, 2], "zx")

    def test_or_1d_swapped(self):
        assert Lattice(1, "x") | Lattice(2) == Lattice([2, 1], "zx")

    def test_or_2d(self):
        assert Lattice([1, 2]) | Lattice(1, "x") == Lattice([1, 2])

    def test_or_2d_swapped(self):
        assert Lattice(1, "x") | Lattice([1, 2]) == Lattice([1, 2])

    def test_or_3d(self):
        assert Lattice([1, 2, 3]) | Lattice(1, "x") == Lattice([1, 2, 3])

    def test_or_3d_swapped(self):
        assert Lattice(2, "y") | Lattice([1, 2, 3]) == Lattice([1, 2, 3])

    def test_or_3d_2d(self):
        assert Lattice([1, 2]) | Lattice([2, 3], "yz") == Lattice([1, 2, 3])

    def test_or_3d_2d_error(self):
        with pytest.raises(ValueError):
            Lattice([1, 2]) | Lattice([3, 3], "yz")

    def test_or_3d_x(self):
        assert Lattice(1, "x") | Lattice([2, 3], "yz") == Lattice([1, 2, 3])

    def test_or_3d_y(self):
        assert Lattice(2, "y") | Lattice([3, 1], "zx") == Lattice([1, 2, 3])

    def test_or_error(self):
        with pytest.raises(ValueError):
            Lattice(1) | Lattice(2)

    def test_and_none(self):
        assert Lattice(1) & None is None

    def test_and_same(self):
        assert Lattice(1) & Lattice(1) == Lattice(1)

    def test_and_empty(self):
        with pytest.raises(ValueError):
            Lattice(1) & Lattice(2)

    def test_and_error(self):
        with pytest.raises(ValueError):
            Lattice(1) & Lattice(1, "x")

    def test_and(self):
        assert Lattice([1, 2, 3]) & Lattice([[0, 1], [3, 0]], "zx") == Lattice(
            [[0, 1], [3, 0]], "zx"
        )

    def test_le(self):
        assert Lattice(3) <= Lattice([1, 2, 3])

    def test_le_false(self):
        assert not Lattice(3) >= Lattice([1, 2, 3])

    def test_isdisjoint(self):
        assert Lattice(1).isdisjoint(Lattice([1, 1]))

    def test_isdisjoint_false(self):
        assert not Lattice(1).isdisjoint(Lattice([1, 1, 1]))
